from ontosunburst.ontosunburst import *
import argparse


def get_command_line_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--input', '-i', type=str, required=True, help='Interest set input')
    parser.add_argument('--ref', '-ir', type=str, required=False, help='Reference set input')
    parser.add_argument('--ontology', '--onto', type=str, required=False, help='Ontology used')
    parser.add_argument('--root', '-r', type=str, required=False, help='Ontology root')
    parser.add_argument('--analysis', '-a', type=str, required=False, default=TOPOLOGY_A,
                        help='Type of analysis')
    parser.add_argument('--output', '-o', type=str, required=False, default='sunburst',
                        help='Output path+name')
    parser.add_argument('--class_ontology', '-cl', type=str, required=False,
                        help='Class ontology json file')
    parser.add_argument('--labels', '-l', type=str, required=False, default=DEFAULT,
                        help='Labels json file')
    parser.add_argument('--url', type=str, required=False, help='Endpoint URL (SPARQL server)')
    parser.add_argument('--test', '-t', type=str, required=False, default=BINOMIAL_TEST,
                        help='Enrichment stat test')
    parser.add_argument('--rcut', type=str, required=False, default=ROOT_CUT,
                        help='Type of root cut')
    parser.add_argument('--pcut', type=str, required=False, default=PATH_UNCUT,
                        help='Type of path cut')
    parser.add_argument('--r_base', action='store_true', required=False, default=False,
                        help='Reference base')
    parser.add_argument('--show_leaves', '-sl',  action='store_true', default=False, required=False,
                        help='Show leaves')
    parser.add_argument('--kwargs', nargs=argparse.REMAINDER, help="Additional keyword arguments")
    args = parser.parse_args()
    return args


def main():
    args = get_command_line_args()
    kwargs = {}
    if args.kwargs:
        for arg in args.kwargs:
            if "=" in arg:
                key, value = arg.split("=", 1)
                kwargs[key] = value
            else:
                raise ValueError(f"Argument {arg} is not in the form key=value")
    reference_set, ref_abundances, scores = extract_input(args.ref)
    metabolic_objects, abundances, scores = extract_input(args.input)
    ontosunburst(interest_set=metabolic_objects,
                 ontology=args.ontology,
                 root=args.root,
                 abundances=abundances,
                 scores=scores,
                 reference_set=reference_set,
                 ref_abundances=ref_abundances,
                 analysis=args.analysis,
                 output=args.output,
                 write_output=True,
                 class_ontology=args.class_ontology,
                 labels=args.labels,
                 endpoint_url=args.url,
                 test=args.test,
                 root_cut=args.rcut,
                 path_cut=args.pcut,
                 ref_base=args.r_base,
                 show_leaves=args.show_leaves,
                 **kwargs)


def extract_input(input_file):
    if input_file is not None:
        id_lst = []
        ab_lst = []
        sc_d = {}
        with open(input_file, 'r') as f:
            for l in f:
                l = l.strip().split('\t')
                id_lst.append(l[0])
                if len(l) == 3:
                    sc_d[l[0]] = float(l[2])
                if len(l) >= 2:
                    ab_lst.append(float(l[1]))
                else:
                    ab_lst.append(1)
        if not sc_d:
            sc_d = None
        return id_lst, ab_lst, sc_d
    return None, None, None



import csv
import json
import os.path

from padmet.utils.sbmlPlugin import convert_from_coded_id
from padmet.classes import PadmetSpec
from typing import List, Set, Tuple


# METABOLITES EXTRACTION
# ==================================================================================================

# From ClusterMap output
def extract_metabolites_clusters(input_file: str, clust: List[int]) -> Set[str]:
    """ Extract metabolites corresponding to clusters desired.

    Parameters
    ----------
    input_file: str
        Cluster tsv file from heatmap output
    clust: List[int]
        List of cluster numbers to extract

    Returns
    -------
    Set[str]
        Set of metabolites corresponding to cluster numbers.
    """
    with open(input_file, 'r') as f:
        rows = list(csv.reader(f, delimiter='\t'))
        del rows[0]
        metabolites = {row[1].replace('_C-BOUNDARY', '') for row in rows if int(row[0]) in clust}
    return metabolites


# From M2M MetaCom output
def extract_community_metabolites_m2m(input_directory: str) -> Tuple[Set[str], Set[str]]:
    """

    Parameters
    ----------
    input_directory: str
        Directory path to m2m results, must contain 'community_analysis' folder

    Returns
    -------
    Tuple[Set[str], Set[str]]
        Set of metabolites from Added Value
        Set of metabolites from Community Scopes
    """
    added_value_json = os.path.join(input_directory, 'community_analysis', 'addedvalue.json')
    comm_scopes_json = os.path.join(input_directory, 'community_analysis', 'comm_scopes.json')
    with open(added_value_json, 'r') as av_f, open(comm_scopes_json, 'r') as cs_f:
        av_set = {convert_from_coded_id(x)[0].replace('_C-BOUNDARY', '') for x in set(json.load(av_f)['addedvalue'])}
        cs_set = {convert_from_coded_id(x)[0].replace('_C-BOUNDARY', '') for x in set(json.load(cs_f)['com_scope'])}
    return av_set, cs_set


def extract_host_metabolites_m2m(input_directory: str) -> Set[str]:
    """

    Parameters
    ----------
    input_directory: str
        Directory path to m2m results, must contain 'community_analysis' folder

    Returns
    -------
    Set[str]
        Set of the host scope metabolites
    """
    comm_scopes_json = os.path.join(input_directory, 'community_analysis', 'comm_scopes.json')
    with open(comm_scopes_json, 'r') as h_f:
        h_set = {convert_from_coded_id(x)[0].replace('_C-BOUNDARY', '') for x in set(json.load(h_f)['host_scope'])}
    return h_set


# From Padmet Network
def extract_network_metabolites(network_file: str, consumes: bool = False) -> Set[str]:
    """

    Parameters
    ----------
    network_file: str
        Path to padmet network file.
    consumes: bool

    Returns
    -------
    Set[str]
        Set of metabolites from the metabolic network.
    """
    padmet = PadmetSpec(network_file)
    test_set = ['produces']
    if consumes:
        test_set.append('consumes')
    me = {rel.id_out for rel in padmet.getAllRelation() if rel.type in test_set}
    me.remove('Bio')
    return me


# From i_scope group selection
def extract_iscope_group_metabolites(scope_json, g_file, group_list):
    with open(scope_json, 'r') as f:
        scopes = json.load(f)
        for sp, met_l in scopes.items():
            scopes[sp] = set(met_l)
    # groups = extract_groups(g_file)
    union_scope = set()
    # for grp in group_list:
    #     sp_list = groups[grp]
    #     all_metabolites = [met for sp, met in scopes.items() if sp in sp_list]
    #     if all_metabolites:
    #         union_scope = union_scope.union(*all_metabolites)
    return set([convert_from_coded_id(met)[0].replace('_C-BOUNDARY', '') for met in union_scope])


# PATHWAYS EXTRACTION
# ==================================================================================================

# From padmet network
# def extract_pathway_classes(padmet_networks_list: List[str], completion_threshold: float = 0) -> Dict[str, List[str]]:
#     """
#
#     Parameters
#     ----------
#     padmet_networks_list: List[str]
#         List of padmet networks path
#     completion_threshold: float (default=0)
#         Minimal completion rate of the pathway to filter.
#
#     Returns
#     -------
#     Dict[str, List[str]]
#         Dict[pathway_ID, List[pathway_class]] : associate for each pathway, its classes associated.
#     """
#     pw_classes = dict()
    # for species_nw in padmet_networks_list:
    #     sp_nw = PadmetNetwork(species_nw)
    #     for pw_id, pw in sp_nw.pathways.items():
    #         if pw.completion_rate > completion_threshold:
    #             pw_classes[pw_id] = pw.is_class
    # return pw_classes


# GO TERMS EXTRACTION
# ==================================================================================================

def extract_go_esmecata(sp_annotations):
    go_abundance = dict()
    for sp_file in sp_annotations:
        with open(sp_file, 'r') as f:
            rows = csv.reader(f, delimiter='\t')
            rows.__next__()
            for row in rows:
                go_terms = row[3].split(',')
                for go in go_terms:
                    if go not in go_abundance:
                        go_abundance[go] = 0
                    go_abundance[go] += 1
    return go_abundance


# EC EXTRACTION
# ==================================================================================================

def extract_ec_esmecata(sp_annotations):
    ec_set = set()
    for sp_file in sp_annotations:
        with open(sp_file, 'r') as f:
            rows = csv.reader(f, delimiter='\t')
            rows.__next__()
            for row in rows:
                ec_number = row[4].split(',')
                for ec in ec_number:
                    if ec != '':
                        ec_set.add(ec)
    return ec_set



import unittest
from unittest.mock import patch
import io
import sys
from functools import wraps
from ontosunburst.ontology import *
from ontosunburst.ontosunburst import *

# ==================================================================================================
# GLOBAL
# ==================================================================================================

GO_LST = ['go:0043227', 'go:0043229', 'go:0043231', 'go:0044422']
GO_URL = 'http://localhost:3030/go/'


# ==================================================================================================
# FUNCTIONS UTILS
# ==================================================================================================
def save_fig_json(fig, file):
    fig = fig.to_dict()
    with open(file, 'w') as f:
        json.dump(fig, f)


def sort_fig_dict_lst(fig):
    """
    {data: [{hovertext: [str],
             ids: [str],
             labels: [str],
             marker: {colors: [int]},
             parents: [str],
             values: [int]}]}
    """
    lst_keys = ['hovertext', 'ids', 'labels', 'parents', 'values']
    for k in lst_keys:
        fig['data'][0][k] = sorted(fig['data'][0][k])
    fig['data'][0]['marker']['colors'] = sorted(fig['data'][0]['marker']['colors'])
    return fig


def fig_to_lines(fig_dict):
    lst_keys = ['hovertext', 'ids', 'labels', 'parents', 'values']
    lines = set()
    for i in range(len(fig_dict['data'][0]['ids'])):
        line = tuple(fig_dict['data'][0][k][i] for k in lst_keys)
        line = line + (str(fig_dict['data'][0]['marker']['colors'][i]),)
        lines.add(line)
    return lines


def are_fig_dict_equals(fig1, fig2_file):
    fig1 = fig1.to_dict()
    fig1_l = fig_to_lines(fig1)
    fig1 = json.dumps(sort_fig_dict_lst(fig1), sort_keys=True)
    with open(fig2_file, 'r') as f:
        fig2 = json.load(f)
        fig2_l = fig_to_lines(fig2)
        fig2 = json.dumps(sort_fig_dict_lst(fig2), sort_keys=True)
    return (fig1 == fig2) and (fig1_l == fig2_l)


def dicts_with_sorted_lists_equal(dict1, dict2):
    if dict1.keys() != dict2.keys():
        return False
    for key in dict1:
        if sorted(dict1[key]) != sorted(dict2[key]):
            return False
    return True


def test_for(func):
    def decorator(test_func):
        @wraps(test_func)
        def wrapper(*args, **kwargs):
            return test_func(*args, **kwargs)
        wrapper._test_for = func
        return wrapper
    return decorator


class DualWriter(io.StringIO):
    def __init__(self, original_stdout):
        super().__init__()
        self.original_stdout = original_stdout

    def write(self, s):
        super().write(s)
        self.original_stdout.write(s)


# ==================================================================================================
# UNIT TESTS
# ==================================================================================================

# TEST ONTOLOGY : CLASSES EXTRACTION
# --------------------------------------------------------------------------------------------------
class TestGOClassesExtraction(unittest.TestCase):
    @test_for(extract_go_classes)
    @patch('sys.stdout', new_callable=lambda: DualWriter(sys.stdout))
    def test_extract_go_classes(self, mock_stdout):
        all_classes, d_classes_ontology, names = extract_go_classes(GO_LST, GO_URL)
        output = mock_stdout.getvalue().strip()
        wanted_ontology = {'go:0043227': ['go:0043226'],
                           'go:0043226': ['go:0110165'],
                           'go:0005575': ['GO'],
                           'go:0110165': ['go:0005575'],
                           'go:0043229': ['go:0043226'],
                           'go:0043231': ['go:0043229', 'go:0043227']}
        wanted_names = {'go:0043227': 'membrane-bounded organelle',
                        'go:0043226': 'organelle',
                        'go:0110165': 'cellular anatomical entity',
                        'go:0005575': 'cellular_component',
                        'go:0043229': 'intracellular organelle',
                        'go:0043231': 'intracellular membrane-bounded organelle'}
        wanted_classes = {'go:0043227': {'go:0043226', 'go:0110165', 'GO', 'go:0005575'},
                          'go:0043229': {'go:0043226', 'go:0110165', 'GO', 'go:0005575'},
                          'go:0043231': {'GO', 'go:0043227', 'go:0043226', 'go:0110165',
                                         'go:0005575', 'go:0043229'}}

        self.assertEqual(output, 'No GO class found for : go:0044422')
        self.assertDictEqual(all_classes, wanted_classes)
        self.assertDictEqual(names, wanted_names)
        self.assertTrue(dicts_with_sorted_lists_equal(d_classes_ontology, wanted_ontology))

    @test_for(extract_classes)
    @patch('sys.stdout', new_callable=lambda: DualWriter(sys.stdout))
    def test_extract_classes_go(self, mock_stdout):
        go_classes, d_classes_ontology, names = extract_classes(GO, GO_LST, ROOTS[GO], None, GO_URL)
        output = mock_stdout.getvalue().strip()
        wanted_ontology = {'go:0043227': ['go:0043226'],
                           'go:0043226': ['go:0110165'],
                           'go:0005575': ['GO'],
                           'go:0110165': ['go:0005575'],
                           'go:0043229': ['go:0043226'],
                           'go:0043231': ['go:0043229', 'go:0043227']}
        wanted_names = {'go:0043227': 'membrane-bounded organelle',
                        'go:0043226': 'organelle',
                        'go:0110165': 'cellular anatomical entity',
                        'go:0005575': 'cellular_component',
                        'go:0043229': 'intracellular organelle',
                        'go:0043231': 'intracellular membrane-bounded organelle'}
        wanted_classes = {'go:0043227': {'go:0043226', 'go:0110165', 'GO', 'go:0005575'},
                          'go:0043229': {'go:0043226', 'go:0110165', 'GO', 'go:0005575'},
                          'go:0043231': {'GO', 'go:0043227', 'go:0043226', 'go:0110165',
                                         'go:0005575', 'go:0043229'}}

        self.assertEqual(output, 'No GO class found for : go:0044422')
        self.assertDictEqual(go_classes, wanted_classes)
        self.assertDictEqual(names, wanted_names)
        self.assertTrue(dicts_with_sorted_lists_equal(d_classes_ontology, wanted_ontology))

    @test_for(ontosunburst)
    def test_ontosunburst_go1(self):
        fig = ontosunburst(interest_set=GO_LST, ontology=GO, root='00',
                           abundances=None, reference_set=None, ref_abundances=None,
                           analysis='topology', output='test_go1', write_output=False,
                           class_ontology=None, labels=None, endpoint_url=None, root_cut='uncut',
                           ref_base=False, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test_go1.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))



from typing import List, Dict, Set
import numpy as np
from numpy import nan
import scipy.stats as stats

# ==================================================================================================
# CONSTANTS
# ==================================================================================================

# Comparison tests
BINOMIAL_TEST = 'binomial'
HYPERGEO_TEST = 'hypergeometric'

# Analysis method
TOPOLOGY_A = 'topology'
ENRICHMENT_A = 'enrichment'

MAX_RELATIVE_NB = 1000000

# Keys
# ----
IDS = 'ID'
ONTO_ID = 'Onto ID'
PARENT = 'Parent'
LABEL = 'Label'
COUNT = 'Count'
REF_COUNT = 'Reference count'
PROP = 'Proportion'
REF_PROP = 'Reference proportion'
RELAT_PROP = 'Relative proportion'
PVAL = 'Pvalue'

# Root cut
ROOT_CUT = 'cut'
ROOT_TOTAL_CUT = 'total'
ROOT_UNCUT = 'uncut'

# Path cut
PATH_UNCUT = 'uncut'
PATH_DEEPER = 'deeper'
PATH_HIGHER = 'higher'
PATH_BOUND = 'bound'


# ==================================================================================================
# CLASS
# ==================================================================================================
class DataTable:
    """
    DataTable class: stores figure parameters values.

    Attributes
    ----------
    self.ids: List[str]
        Unique ids of all sectors (id=path in the tree)
    self.onto_ids: List[str]
        Ontology ID of all sectors
    self.labels: List[str]
        Sectors labels
    self.parents: List[str]
        Sectors parents
    self.count: List[float]
        Sectors interest set count (0<count)
    self.ref_count: List[float]
        Sectors reference set count (0<count)
    self.prop: List[float]
        Sectors interest set proportion (0<prop<1)
    self.ref_prop: List[float]
        Sectors reference set proportion (0<prop<1)
    self.relative_prop: List[int]
        Sectors relative proportion (0<r_prop<1000000)
    self.p_val: List[float]
        Sectors p-value if enrichment analysis
    self.len: int
        Number of sectors
    """
    def __init__(self):
        self.ids = list()
        self.onto_ids = list()
        self.labels = list()
        self.parents = list()
        self.count = list()
        self.ref_count = list()
        self.prop = list()
        self.ref_prop = list()
        self.relative_prop = list()
        self.p_val = list()
        self.len = 0

    def __str__(self):
        string = ''
        data = self.get_data_dict()
        for k, v in data.items():
            string += f'{k}\n{"-"*len(k)}\n{v}\n'
        return string

    def get_data_dict(self):
        return {IDS: self.ids, ONTO_ID: self.onto_ids, LABEL: self.labels, PARENT: self.parents,
                COUNT: self.count, REF_COUNT: self.ref_count, PROP: self.prop,
                REF_PROP: self.ref_prop, RELAT_PROP: self.relative_prop, PVAL: self.p_val}

    def fill_parameters(self, set_abundance: Dict[str, float], ref_abundance: Dict[str, float],
                        parent_dict: Dict[str, List[str]], root_item: str,
                        names: Dict[str, str] = None, ref_base: bool = True):
        """ Fill DataTable list attributes (self.ids, self.onto_ids, self.labels, self.parents,
        self.count, self.ref_count)

        Parameters
        ----------
        set_abundance: Dict[str, float]
            Dictionary associating for each class the number of objects found belonging to the class
            in the interest set
        ref_abundance: Dict[str, float]
            Dictionary associating for each class the number of objects found belonging to the class
            in the reference set
        parent_dict: Dict[str, List[str]]
            Dictionary associating for each class, its parents classes
        root_item: str
            Name of the root item of the ontology
        names: Dict[str, str]
            Dictionary associating for some or each ontology IDs, its label
        ref_base: bool
            True to have the reference as base, False otherwise
        """
        if ref_base:
            for c_onto_id, c_ref_abundance in ref_abundance.items():
                c_abundance = get_set2_abundance(set_abundance, c_onto_id)
                self.__fill_id_parameter(c_onto_id, root_item, names, parent_dict, c_abundance,
                                         c_ref_abundance)
        else:
            for c_onto_id, c_abundance in set_abundance.items():
                c_ref_abundance = get_set2_abundance(ref_abundance, c_onto_id)
                self.__fill_id_parameter(c_onto_id, root_item, names, parent_dict, c_abundance,
                                         c_ref_abundance)

    def __fill_id_parameter(self, c_onto_id: str, root_item: str, names: Dict[str, str],
                            parent_dict: Dict[str, List[str]], c_abundance: float,
                            c_ref_abundance: float):
        """ Fill DataTable list attributes (self.ids, self.onto_ids, self.labels, self.parents,
        self.count, self.ref_count) for one concept.

        Parameters
        ----------
        c_onto_id: str
            Ontology id of the concept
        root_item: str
            Root of the ontology
        names: Dict[str, str]
            Dictionary associating for some or each ontology IDs, its label
        parent_dict: Dict[str: List[str]]
            Dictionary associating for each class, its parents classes
        c_abundance: float
            Abundance of the concept in the interest set
        c_ref_abundance: float
            Abundance of the concept in the reference set
        """
        if c_onto_id != root_item:
            c_label = get_name(c_onto_id, names)
            all_c_ids = get_all_ids(c_onto_id, c_onto_id, parent_dict, root_item, set())
            for c_id in all_c_ids:
                self.add_value(m_id=c_id, onto_id=c_onto_id, label=c_label,
                               count=c_abundance, ref_count=c_ref_abundance,
                               parent=c_id[len(c_onto_id) + 2:])  # Remove c_label__ prefix
        else:
            self.add_value(m_id=c_onto_id, onto_id=c_onto_id, label=c_onto_id,
                           count=c_abundance, ref_count=c_ref_abundance, parent='')

    def add_value(self, m_id: str, onto_id: str, label: str, count: float, ref_count: float,
                  parent: str):
        """ Fill the data attributes for an object class.

        Parameters
        ----------
        m_id: str
            ID unique of the object class to add
        onto_id: str
            ID in the ontology
        label: str
            Label (name) of the object class to add
        count: int
            Abundance value of the object class to add
        ref_count: int
            Reference abundance value of the object class to add
        parent: str
            Parent object class of the object class to add
        """
        if m_id in self.ids:
            raise ValueError(f'{m_id} already in data IDs, all IDs must be unique.')
        self.ids.append(m_id)
        self.onto_ids.append(onto_id)
        self.labels.append(label)
        self.parents.append(parent)
        self.count.append(count)
        self.ref_count.append(ref_count)
        self.prop.append(nan)
        self.ref_prop.append(nan)
        self.relative_prop.append(nan)
        self.p_val.append(nan)
        self.len += 1

    def calculate_proportions(self, ref_base: bool):
        """ Calculate DataTable proportion list attributes (self.prop, self.ref_prop,
        self.relative_prop). If total add relative proportion to +1 parent for branch value.

        Parameters
        ----------
        ref_base: bool
            True if reference base representation
        """
        # Get total proportion
        max_abondance = int(np.nanmax(self.count))
        self.prop = [x / max_abondance for x in self.count]
        # Get reference proportion
        max_ref_abondance = np.max(self.ref_count)
        self.ref_prop = [x / max_ref_abondance for x in self.ref_count]
        # Get proportion relative to +1 parent proportion for total branch value
        self.relative_prop = [x for x in self.prop]
        p = ''
        self.__get_relative_prop(p, ref_base)
        # IDK WHY IT WORKS ???
        missed = [self.ids[i] for i in range(self.len) if self.relative_prop[i] < 1]
        if missed:
            parents = {self.parents[self.ids.index(m)] for m in missed}
            for p in parents:
                self.__get_relative_prop(p, ref_base)

    def __get_relative_prop(self, p_id: str, ref_base: bool):
        """ Get recursively relative proportion of a parent children to itself. Set it to class
        self.relative_prop attribute.

        Parameters
        ----------
        p_id: str
            ID of the parent
        ref_base: bool
            True if reference base representation
        """
        if ref_base:
            base_count = self.ref_count
        else:
            base_count = self.count
        if p_id == '':
            prop_p = MAX_RELATIVE_NB
            count_p = max(base_count)
        else:
            prop_p = self.relative_prop[self.ids.index(p_id)]
            count_p = base_count[self.ids.index(p_id)]
        index_p = [i for i, v in enumerate(self.parents) if v == p_id]
        p_children = [self.ids[i] for i in index_p]
        count_p_children = [base_count[i] for i in index_p]
        if np.nansum(count_p_children) > count_p:
            total = np.nansum(count_p_children)
        else:
            total = count_p
        for i, c in enumerate(p_children):
            if not ref_base and np.isnan(self.prop[self.ids.index(c)]):
                prop_c = 0
            else:
                prop_c = int((count_p_children[i] / total) * prop_p)
            self.relative_prop[self.ids.index(c)] = prop_c
        for c in p_children:
            if c in self.parents:
                self.__get_relative_prop(c, ref_base)

    def make_enrichment_analysis(self, test: str, scores: Dict[str, float] = None) \
            -> Dict[str, float]:
        """ Performs statistical tests for enrichment analysis.

        Parameters
        ----------
        test: str
            Type of test : binomial or hypergeometric
        scores: Dict[str, float]
            Dictionary associating for each ontology ID, its enrichment score. If None enrichment
            will be calculated.

        Returns
        -------
        Dict[str, float]
            Dictionary of significant metabolic object label associated with their p-value
        """
        nb_classes = len(set([self.labels[i] for i in range(self.len)
                              if not np.isnan(self.count[i])]))
        significant_representation = dict()
        if scores is not None:
            for i in range(self.len):
                p_val = scores[self.onto_ids[i]]
                self.p_val[i] = -np.log10(p_val)
                if p_val < 0.05 / nb_classes:  # Keep significant p-values : Bonferroni
                    significant_representation[self.onto_ids[i]] = p_val
        else:
            m = np.max(self.ref_count)  # M = ref set total item number
            n = int(np.nanmax(self.count))  # N = interest set total item number
            for i in range(self.len):
                if type(self.count[i]) == int:  # If count not nan (= if concept in interest set)
                    # Binomial Test
                    if test == BINOMIAL_TEST:
                        p_val = stats.binomtest(self.count[i], n, self.ref_count[i] / m,
                                                alternative='two-sided').pvalue
                    # Hypergeometric Test
                    elif test == HYPERGEO_TEST:
                        p_val_upper = stats.hypergeom.sf(self.count[i] - 1, m, self.ref_count[i], n)
                        p_val_lower = stats.hypergeom.cdf(self.count[i], m, self.ref_count[i], n)
                        p_val = 2 * min(p_val_lower, p_val_upper)  # bilateral
                    else:
                        raise ValueError(
                            f'test parameter must be in : {[BINOMIAL_TEST, HYPERGEO_TEST]}')
                    if ((self.count[i] / n) - (self.ref_count[i] / m)) > 0:  # If over-represented :
                        self.p_val[i] = -np.log10(p_val)  # Positive log10(p-value)
                    else:  # If under-represented :
                        self.p_val[i] = np.log10(p_val)  # Negative log10(p-value)
                    if p_val < 0.05 / nb_classes:  # Keep significant p-values : Bonferroni
                        significant_representation[self.onto_ids[i]] = p_val
        significant_representation = dict(
            sorted(significant_representation.items(), key=lambda item: item[1]))
        return significant_representation

    def cut_root(self, mode: str):
        """ Filter data to cut (or not) the root to remove not necessary 100% represented classes.

        Parameters
        ----------
        mode: str
            Mode of root cutting
            - uncut: doesn't cut and keep all nodes from ontology root
            - cut: keep only the lowest level 100% shared node
            - total: remove all 100% shared nodes (produces a pie at center)
        """
        if mode not in {ROOT_UNCUT, ROOT_CUT, ROOT_TOTAL_CUT}:
            raise ValueError(f'Root cutting mode {mode} unknown, '
                             f'must be in {[ROOT_UNCUT, ROOT_CUT, ROOT_TOTAL_CUT]}')
        if mode == ROOT_CUT or mode == ROOT_TOTAL_CUT:
            roots_ind = [i for i in range(self.len) if self.relative_prop[i] == MAX_RELATIVE_NB]
            roots = [self.ids[i] for i in roots_ind]
            self.delete_value(roots_ind)
            if mode == ROOT_CUT:
                self.parents = [str(x).split('__')[0] if x in roots else x for x in self.parents]
            if mode == ROOT_TOTAL_CUT:
                self.parents = ['' if x in roots else x for x in self.parents]

    def cut_nested_path(self, mode: str, ref_base: bool):
        """ Cut nested path in the tree graph (path of nested sectors sharing the same value)

        Parameters
        ----------
        mode: str
            Mode of path cutting
            - uncut: doesn't cut and keep all sectors
            - deeper: cut nested path and only conserve the deepest sector in the tree
            - higher: cut nested path and only conserve the highest sector in the tree
            - bound: cut nested path and only conserve the highest AND the deepest sectors in the
            tree
        ref_base: bool
            True if reference base representation
        """
        if ref_base:
            count = self.ref_count
        else:
            count = self.count
        if mode != PATH_UNCUT:
            nested_paths = []
            for p_i in range(self.len):
                p = self.ids[p_i]
                p_children = [self.ids[i] for i in range(self.len) if self.parents[i] == p]
                if len(p_children) == 1:
                    p_p = self.parents[p_i]
                    p_p_children = [self.ids[i] for i in range(self.len) if self.parents[i] == p_p]
                    if len(p_p_children) != 1:
                        p_count = count[p_i]
                        c_i = self.ids.index(p_children[0])
                        c_count = count[c_i]
                        if p_count == c_count:
                            nested_paths.append(self.get_full_nested_path(c_i, [p_i], count))
            self.delete_nested_path(mode, nested_paths)

    def get_full_nested_path(self, p_i: int, n_path: List[int], count: List[float]):
        """ Get all index of a nested path sector from its parent sector index.

        Parameters
        ----------
        p_i: int
            Parent sector index of the nested path
        n_path: List[int]
            List of sector indexes of the nested path
        count: List[float]
            List of all sectors count value.

        Returns
        -------
        List[int]
            List of sector indexes of the nested path
        """
        n_path.append(p_i)
        p = self.ids[p_i]
        p_children = [self.ids[i] for i in range(self.len) if self.parents[i] == p]
        if len(p_children) == 1:
            p_count = count[p_i]
            c_i = self.ids.index(p_children[0])
            c_count = count[c_i]
            if p_count == c_count:
                n_path = self.get_full_nested_path(c_i, n_path, count)
        return n_path

    def delete_nested_path(self, mode: str, nested_paths: List[List[int]]):
        """ Delete some sectors of the nested path to conserve only the deepest (deeper mode), only
        the highest (higher mode) or both (bound mode)

        Parameters
        ----------
        mode: str
            Mode of path cutting
            - uncut: doesn't cut and keep all sectors
            - deeper: cut nested path and only conserve the deepest sector in the tree
            - higher: cut nested path and only conserve the highest sector in the tree
            - bound: cut nested path and only conserve the highest AND the deepest sectors in the
            tree
        nested_paths: List[List[int]]
            List of lists of nested path sectors indexes
        """
        to_del = []
        if mode == PATH_DEEPER:
            for path in nested_paths:
                to_del += path[:-1]
                to_keep = path[-1]
                root_p = self.parents[path[0]]
                self.parents[to_keep] = root_p
                self.labels[to_keep] = '... ' + self.labels[to_keep]
        elif mode == PATH_HIGHER:
            for path in nested_paths:
                to_del += path[1:]
                to_keep = path[0]
                to_keep_c = [i for i in range(self.len) if self.parents[i] == self.ids[path[-1]]]
                for c_i in to_keep_c:
                    self.parents[c_i] = self.ids[to_keep]
                self.labels[to_keep] += ' ...'
        elif mode == PATH_BOUND:
            for path in nested_paths:
                to_del += path[1:-1]
                to_keep_up = path[0]
                to_keep_do = path[-1]
                self.parents[to_keep_do] = self.ids[to_keep_up]
                if len(nested_paths) > 2:
                    self.labels[to_keep_up] += ' ...'
                    self.labels[to_keep_do] = '... ' + self.labels[to_keep_do]
        self.delete_value(to_del)

    def delete_value(self, v_index: int or List[int]):
        """ Delete a sector of DataTable from its index or a list of sectors from a list of indexes

        Parameters
        ----------
        v_index: int or List[int]
            Index or list of indexes of the sectors to delete
        """
        data = self.get_data_dict()
        if type(v_index) == int:
            v_index = [v_index]
        for i in sorted(v_index, reverse=True):
            for k, v in data.items():
                del v[i]
            self.len -= 1

    def get_col(self, index: int or List[int] = None) -> List or List[List]:
        """ Get a DataTable column from its index or a list of columns from a list of indexes.
        Column = all values associated with a sector.

        Parameters
        ----------
        index: int or List[int]
            Index or list of indexes of the sectors to get the column

        Returns
        -------
        List or List[List]
            Column or list of columns obtained from indexes
        """
        if index is None:
            index = list(range(self.len))
        if type(index) == int:
            index = [index]
        cols = list()
        for i in index:
            cols.append((self.ids[i], self.onto_ids[i], self.labels[i], self.parents[i],
                         self.count[i], self.ref_count[i], self.prop[i], self.ref_prop[i],
                         self.relative_prop[i], self.p_val[i]))
        return cols


# ==================================================================================================
# FUNCTIONS
# ==================================================================================================
def get_all_ids(m_id: str, n_id: str, parent_dict: Dict[str, List[str]], root: str,
                all_ids: Set[str]) -> Set[str]:
    """ Return recursively all unique IDs associated with a label. The IDs correspond to the path
    in the tree from the label to the root.

    Parameters
    ----------
    m_id: str
        Input ID
    n_id
        New ID
    parent_dict: Dict[str, List[str]]
        Dictionary associating for each class, its parents classes
    root: str
        Name of the root item of the ontology
    all_ids: Set[str]
        Set of unique IDs associated with a concept label.

    Returns
    -------
    Set[str]
        Set of all unique IDs associated with a concept label.
    """
    parents = parent_dict[m_id]
    for p in parents:
        nn_id = n_id + '__' + p
        if p == root:
            all_ids.add(nn_id)
        else:
            all_ids = get_all_ids(p, nn_id, parent_dict, root, all_ids)
    return all_ids


def get_set2_abundance(set2_abundances: Dict[str, float] or None, c_label: str) -> float:
    """ Get the set2 abundance of a set1 concept.

    Parameters
    ----------
    set2_abundances: Dict[str, float] or None
        Dictionary associating to all the set1 concepts, its abundance
    c_label: str
        Label of the set2 concept

    Returns
    -------
    Abundance of the set1 concept. If set2 concept not in the set1, returns numpy.nan.
    """
    try:
        c_set2_abundance = set2_abundances[c_label]
    except KeyError:
        c_set2_abundance = np.nan
    return c_set2_abundance


def get_name(c_onto_id, names):
    if names is not None:
        try:
            c_label = names[c_onto_id]
        except KeyError:
            c_label = c_onto_id
    else:
        c_label = c_onto_id
    return c_label



import json
import os.path
import unittest
import io
import sys
import copy
from functools import wraps
from unittest.mock import patch

from ontosunburst.ontology import ROOTS, METACYC
from ontosunburst.sunburst_fig import *

"""
Tests manually good file creation.
No automatic tests integrated.
"""

# ==================================================================================================
# GLOBAL
# ==================================================================================================

# Complex topology
# --------------------------------------------------------------------------------------------------
MC_ONTO = {'a': ['ab'], 'b': ['ab'], 'c': ['cde', 'cf'], 'd': ['cde'], 'e': ['cde', 'eg'],
           'f': ['cf'], 'g': ['gh', 'eg'], 'h': ['gh'],
           'ab': [ROOTS[METACYC]], 'cde': ['cdecf', 'cdeeg'], 'cf': ['cdecf'],
           'eg': ['cdeeg', ROOTS[METACYC]], 'gh': [ROOTS[METACYC]],
           'cdecf': [ROOTS[METACYC]], 'cdeeg': ['cdeeg+'], 'cdeeg+': [ROOTS[METACYC]]}

MC_AB = {'FRAMES': 6, 'cde': 3, 'cf': 3, 'cdecf': 3, 'cdeeg+': 3, 'cdeeg': 3, 'c': 3, 'ab': 3,
         'b': 2, 'a': 1}
MC_REF_AB = {'FRAMES': 36, 'cdeeg+': 19, 'cdeeg': 19, 'cdecf': 18, 'gh': 15, 'eg': 12, 'cde': 12,
             'cf': 9, 'h': 8, 'g': 7, 'f': 6, 'e': 5, 'd': 4, 'c': 3, 'ab': 3, 'b': 2, 'a': 1}

MC_LABELS = {'FRAMES': 'Root', 'cdeeg+': 'CDEEG+', 'cdeeg': 'CDEEG', 'cdecf': 'CDECF', 'gh': 'GH',
             'eg': 'EG', 'cde': 'CDE', 'cf': 'CF', 'h': 'H', 'g': 'G', 'f': 'F', 'e': 'E', 'd': 'D',
             'c': 'C', 'ab': 'AB', 'b': 'B'}
MC_DATA = DataTable()
MC_DATA.fill_parameters(MC_AB, MC_REF_AB, MC_ONTO, ROOTS[METACYC], MC_LABELS)
MC_DATA.calculate_proportions(True)

# Enrichment
E_AB = {'00': 50, '01': 5, '02': 25, '03': 20, '04': 1, '05': 5, '06': nan, '07': nan,
        '08': 1, '09': 1}
E_REF_AB = {'00': 100, '01': 40, '02': 30, '03': 20, '04': 10, '05': 20, '06': 5, '07': 1,
            '08': 1, '09': 3}
E_LABElS = {'00': '0', '01': '1', '02': '2', '03': '3', '04': '4',
            '05': '5', '06': '6', '07': '7', '08': '8', '09': '9'}
E_ONTO = {'01': ['00'], '02': ['00'], '03': ['00'], '04': ['00'], '05': ['01'],
          '06': ['01'], '07': ['01'], '08': ['02'], '09': ['02']}

E_DATA = DataTable()
E_DATA.fill_parameters(E_AB, E_REF_AB, E_ONTO, '00', E_LABElS, True)
E_DATA.calculate_proportions(True)
E_SIGN = E_DATA.make_enrichment_analysis(BINOMIAL_TEST)


# ==================================================================================================
# FUNCTIONS UTILS
# ==================================================================================================
def dicts_with_sorted_lists_equal(dict1, dict2):
    if dict1.keys() != dict2.keys():
        return False
    for key in dict1:
        if sorted(dict1[key]) != sorted(dict2[key]):
            return False
    return True


def save_fig_json(fig, file):
    fig = fig.to_dict()
    with open(file, 'w') as f:
        json.dump(fig, f)


def sort_fig_dict_lst(fig):
    """
    {data: [{hovertext: [str],
             ids: [str],
             labels: [str],
             marker: {colors: [int]},
             parents: [str],
             values: [int]}]}
    """
    lst_keys = ['hovertext', 'ids', 'labels', 'parents', 'values']
    for k in lst_keys:
        fig['data'][0][k] = sorted(fig['data'][0][k])
    fig['data'][0]['marker']['colors'] = sorted(fig['data'][0]['marker']['colors'])
    return fig


def fig_to_lines(fig_dict):
    lst_keys = ['hovertext', 'ids', 'labels', 'parents', 'values']
    lines = set()
    for i in range(len(fig_dict['data'][0]['ids'])):
        line = tuple(fig_dict['data'][0][k][i] for k in lst_keys)
        line = line + (str(fig_dict['data'][0]['marker']['colors'][i]),)
        lines.add(line)
    return lines


def are_fig_dict_equals(fig1, fig2_file):
    fig1 = fig1.to_dict()
    fig1_l = fig_to_lines(fig1)
    fig1 = json.dumps(sort_fig_dict_lst(fig1), sort_keys=True)
    with open(fig2_file, 'r') as f:
        fig2 = json.load(f)
        fig2_l = fig_to_lines(fig2)
        fig2 = json.dumps(sort_fig_dict_lst(fig2), sort_keys=True)
    return (fig1 == fig2) and (fig1_l == fig2_l)


def test_for(func):
    def decorator(test_func):
        @wraps(test_func)
        def wrapper(*args, **kwargs):
            return test_func(*args, **kwargs)

        wrapper._test_for = func
        return wrapper

    return decorator


class DualWriter(io.StringIO):
    def __init__(self, original_stdout):
        super().__init__()
        self.original_stdout = original_stdout

    def write(self, s):
        super().write(s)
        self.original_stdout.write(s)


# ==================================================================================================
# UNIT TESTS
# ==================================================================================================

# TEST
# --------------------------------------------------------------------------------------------------

class TestSunburstFigure(unittest.TestCase):
    @test_for(get_fig_kwargs)
    def test_get_kwargs_topology(self):
        c_min, c_max, c_mid, max_depth, colorscale, title, colorbar_legend, background_color, \
            font_color, font_size, table_title, table_legend, table_color = \
            get_fig_kwargs('out', TOPOLOGY_A)
        self.assertEqual(c_min, 1)
        self.assertEqual(c_max, None)
        self.assertEqual(c_mid, None)
        self.assertEqual(max_depth, 7)
        self.assertEqual(colorscale, px.colors.get_colorscale('Viridis'))
        self.assertEqual(title, 'out : Proportion of classes')
        self.assertEqual(colorbar_legend, 'Count')
        self.assertEqual(background_color, 'rgba(255, 255, 255, 0)')
        self.assertEqual(font_color, '#111111')
        self.assertEqual(font_size, 20)
        self.assertEqual(table_title, 'Significant p-values')
        self.assertEqual(table_legend, 'IDs')
        self.assertEqual(table_color, '#666666')

    @test_for(get_fig_kwargs)
    def test_get_kwargs_enrichment(self):
        c_min, c_max, c_mid, max_depth, colorscale, title, colorbar_legend, background_color, \
            font_color, font_size, table_title, table_legend, table_color = \
            get_fig_kwargs('out', ENRICHMENT_A)
        self.assertEqual(c_min, -10)
        self.assertEqual(c_max, 10)
        self.assertEqual(c_mid, 0)
        self.assertEqual(max_depth, 7)
        self.assertEqual(colorscale, px.colors.get_colorscale('RdBu'))
        self.assertEqual(title, 'out : Classes enrichment representation')
        self.assertEqual(colorbar_legend, 'Log10(p-value)')
        self.assertEqual(background_color, 'rgba(255, 255, 255, 0)')
        self.assertEqual(font_color, '#111111')
        self.assertEqual(font_size, 20)
        self.assertEqual(table_title, 'Significant p-values')
        self.assertEqual(table_legend, 'IDs')
        self.assertEqual(table_color, '#666666')

    @test_for(get_fig_kwargs)
    def test_get_kwargs(self):
        c_min, c_max, c_mid, max_depth, colorscale, title, colorbar_legend, background_color, \
            font_color, font_size, table_title, table_legend, table_color = \
            get_fig_kwargs('out', TOPOLOGY_A, c_min=4, c_max=8, c_mid=6, max_depth=10,
                           colorscale='Twilight', title='My title', colorbar_legend='Total',
                           bg_color='#000000', font_color='#ffffff', font_size=24,
                           table_title='Significant classes', table_legend='Name',
                           table_color='#222222')
        self.assertEqual(c_min, 4)
        self.assertEqual(c_max, 8)
        self.assertEqual(c_mid, 6)
        self.assertEqual(max_depth, 10)
        self.assertEqual(colorscale, px.colors.get_colorscale('Twilight'))
        self.assertEqual(title, 'My title')
        self.assertEqual(colorbar_legend, 'Total')
        self.assertEqual(background_color, '#000000')
        self.assertEqual(font_color, '#ffffff')
        self.assertEqual(font_size, 24)
        self.assertEqual(table_title, 'Significant classes')
        self.assertEqual(table_legend, 'Name')
        self.assertEqual(table_color, '#222222')

    @test_for(check_kwargs)
    @patch('sys.stdout', new_callable=lambda: DualWriter(sys.stdout))
    def test_check_kwargs(self, mock_stdout):
        check_kwargs(bg_color='black', cmin=2, max_detph=3, c_mif=10, framboise='fruit',
                     c_max='blue', backgroung_color='white')
        output = mock_stdout.getvalue().strip()
        expected_msg = 'Unknown kwarg "cmin", did you mean "c_min" ?\n' \
                       'Unknown kwarg "max_detph", did you mean "max_depth" ?\n' \
                       'Unknown kwarg "c_mif", did you mean "c_min" ?\n' \
                       'Unknown kwarg "framboise"\n' \
                       '"c_max" must be of type "<class \'float\'>" not "<class \'str\'>"\n' \
                       'Unknown kwarg "backgroung_color", did you mean "bg_color" ?'
        self.assertEqual(output, expected_msg)

    @test_for(get_hover_fig_text)
    def test_get_hover_fig_text_enrich_ref(self):
        data = copy.deepcopy(E_DATA)
        text_list = get_hover_fig_text(data, ENRICHMENT_A, True)
        self.assertEqual(len(text_list), 10)
        self.assertEqual(text_list[0], 'P value: 1.0<br>Count: <b>50</b>'
                                       '<br>Reference count: 100<br>Proportion: <b>100.0%</b>'
                                       '<br>Reference proportion: 100.0%<br>ID: 00')
        self.assertEqual(text_list[5], 'P value: 0.07883064215278136<br>Count: <b>5</b>'
                                       '<br>Reference count: 20<br>Proportion: <b>10.0%</b>'
                                       '<br>Reference proportion: 20.0%<br>ID: 05')

    @test_for(get_hover_fig_text)
    def test_get_hover_fig_text_enrich_no_ref(self):
        data = copy.deepcopy(E_DATA)
        text_list = get_hover_fig_text(data, ENRICHMENT_A, False)
        self.assertEqual(len(text_list), 10)
        self.assertEqual(text_list[0], 'P value: 1.0<br>Count: <b>50</b><br>'
                                       'Reference count: 100<br>Proportion: <b>100.0%</b><br>'
                                       'Reference proportion: 100.0%<br>ID: 00')
        self.assertEqual(text_list[5], 'P value: 0.07883064215278136<br>Count: <b>5</b><br>'
                                       'Reference count: 20<br>Proportion: <b>10.0%</b><br>'
                                       'Reference proportion: 20.0%<br>ID: 05')

    @test_for(get_hover_fig_text)
    def test_get_hover_fig_text_topology_ref(self):
        data = copy.deepcopy(E_DATA)
        text_list = get_hover_fig_text(data, TOPOLOGY_A, True)
        self.assertEqual(len(text_list), 10)
        self.assertEqual(text_list[0], 'Count: <b>50</b><br>Reference count: 100'
                                       '<br>Proportion: <b>100.0%</b>'
                                       '<br>Reference proportion: 100.0%<br>ID: 00')
        self.assertEqual(text_list[5], 'Count: <b>5</b><br>Reference count: 20'
                                       '<br>Proportion: <b>10.0%</b>'
                                       '<br>Reference proportion: 20.0%<br>ID: 05')

    @test_for(get_hover_fig_text)
    def test_get_hover_fig_text_topology_no_ref(self):
        data = copy.deepcopy(E_DATA)
        text_list = get_hover_fig_text(data, TOPOLOGY_A, False)
        self.assertEqual(len(text_list), 10)
        self.assertEqual(text_list[0], 'Count: <b>50</b><br>Proportion: <b>100.0%</b><br>ID: 00')
        self.assertEqual(text_list[5], 'Count: <b>5</b><br>Proportion: <b>10.0%</b><br>ID: 05')

    @test_for(generate_sunburst_fig)
    def test_generate_sunburst_fig_case1(self):
        data = copy.deepcopy(E_DATA)
        fig = generate_sunburst_fig(data, 'case1', analysis=ENRICHMENT_A, write_fig=False,
                                    test=HYPERGEO_TEST, ref_set=True, significant=E_SIGN)
        w_fig_file = os.path.join('test_files', 'fig_case1.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    @test_for(generate_sunburst_fig)
    def test_generate_sunburst_fig_case2(self):
        data = copy.deepcopy(E_DATA)
        fig = generate_sunburst_fig(data, 'case2', analysis=ENRICHMENT_A, write_fig=False,
                                    test=BINOMIAL_TEST, significant=E_SIGN)
        w_fig_file = os.path.join('test_files', 'fig_case2.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    @test_for(generate_sunburst_fig)
    def test_generate_sunburst_fig_case3(self):
        data = copy.deepcopy(E_DATA)
        fig = generate_sunburst_fig(data, 'case3', analysis=ENRICHMENT_A,
                                    test=HYPERGEO_TEST, significant=E_SIGN,
                                    root_cut=ROOT_UNCUT, write_fig=False,
                                    title='Another title', colorscale='PuOr_r',
                                    bg_color='#222222', font_color='#eeeeee', font_size=25,
                                    table_legend='Number')
        w_fig_file = os.path.join('test_files', 'fig_case3.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    @test_for(generate_sunburst_fig)
    def test_generate_sunburst_fig_case4(self):
        data = copy.deepcopy(MC_DATA)
        fig = generate_sunburst_fig(data, 'case4', analysis=TOPOLOGY_A, bg_color='black',
                                    font_color='white', write_fig=False)
        w_fig_file = os.path.join('test_files', 'fig_case4.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))



import os.path
import difflib
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from ontosunburst.data_table_tree import *

# ==================================================================================================
# CONSTANTS
# ==================================================================================================

# Kwargs
C_MIN = 'c_min'
C_MAX = 'c_max'
C_MID = 'c_mid'
MAX_DEPTH = 'max_depth'
COLORSCALE = 'colorscale'
TITLE = 'title'
COLORBAR_LEGEND = 'colorbar_legend'
BG_COLOR = 'bg_color'
FONT_COLOR = 'font_color'
FONT_SIZE = 'font_size'
TABLE_TITLE = 'table_title'
TABLE_LEGEND = 'table_legend'
TABLE_COLOR = 'table_color'
KWARGS = [C_MIN, C_MAX, C_MID, MAX_DEPTH, COLORSCALE, TITLE, COLORBAR_LEGEND, BG_COLOR, FONT_COLOR,
          FONT_SIZE, TABLE_TITLE, TABLE_LEGEND, TABLE_COLOR]
KWARGS_TYPE = {C_MIN: float, C_MAX: float, C_MID: float, MAX_DEPTH: int, COLORSCALE: str,
               TITLE: str, COLORBAR_LEGEND: str, BG_COLOR: str, FONT_COLOR: str, FONT_SIZE: int,
               TABLE_TITLE: str, TABLE_LEGEND: str, TABLE_COLOR: str}


# ==================================================================================================
# FUNCTIONS
# ==================================================================================================

# Figure creation
# --------------------------------------------------------------------------------------------------
def get_fig_kwargs(output: str, analysis: str, **kwargs):
    """ Generate a Sunburst figure and save it to output path.

        Parameters
        ----------
        output: str (optional, default=None)
            Path to output to save the figure without extension
        analysis: str (optional, default=topology)
            Analysis mode : topology or enrichment
        """
    check_kwargs(**kwargs)
    def_colorscale = {TOPOLOGY_A: 'Viridis',
                      ENRICHMENT_A: 'RdBu'}
    def_titles = {TOPOLOGY_A: f'{os.path.basename(output)} : Proportion of classes',
                  ENRICHMENT_A: f'{os.path.basename(output)} : Classes enrichment representation'}
    def_colorbar = {TOPOLOGY_A: 'Count',
                    ENRICHMENT_A: 'Log10(p-value)'}
    def_c_min = {TOPOLOGY_A: 1, ENRICHMENT_A: -10}
    def_c_max = {TOPOLOGY_A: None, ENRICHMENT_A: 10}
    def_c_mid = {TOPOLOGY_A: None, ENRICHMENT_A: 0}

    c_min = kwargs.get(C_MIN, def_c_min[analysis])
    c_max = kwargs.get(C_MAX, def_c_max[analysis])
    c_mid = kwargs.get(C_MID, def_c_mid[analysis])
    max_depth = kwargs.get(MAX_DEPTH, 7)
    colorscale = px.colors.get_colorscale(kwargs.get(COLORSCALE, def_colorscale[analysis]))
    title = kwargs.get(TITLE, def_titles[analysis])
    colorbar_legend = kwargs.get(COLORBAR_LEGEND, def_colorbar[analysis])
    background_color = kwargs.get(BG_COLOR, 'rgba(255, 255, 255, 0)')
    font_color = kwargs.get(FONT_COLOR, '#111111')
    font_size = kwargs.get(FONT_SIZE, 20)
    table_title = kwargs.get(TABLE_TITLE, 'Significant p-values')
    table_legend = kwargs.get(TABLE_LEGEND, 'IDs')
    table_color = kwargs.get(TABLE_COLOR, '#666666')

    return c_min, c_max, c_mid, max_depth, colorscale, title, colorbar_legend, background_color, \
        font_color, font_size, table_title, table_legend, table_color


def check_kwargs(**kwargs):
    close_matches = {x: difflib.get_close_matches(x, KWARGS, n=1, cutoff=0.5)[0] for x in kwargs
                     if difflib.get_close_matches(x, KWARGS, n=1, cutoff=0.5) and x not in KWARGS}
    for k in kwargs:
        if k not in KWARGS:
            if k in close_matches:
                print(f'Unknown kwarg "{k}", did you mean "{close_matches[k]}" ?')
            else:
                print(f'Unknown kwarg "{k}"')
        elif type(k) != KWARGS_TYPE[k]:
            print(f'"{k}" must be of type "{KWARGS_TYPE[k]}" not "{type(k)}"')


def generate_sunburst_fig(data: DataTable, output: str, analysis: str = TOPOLOGY_A,
                          test=BINOMIAL_TEST, significant: Dict[str, float] = None,
                          ref_set: bool = True, write_fig: bool = True, **kwargs) -> go.Figure:
    """ Generate a Sunburst figure and save it to output path.

    Parameters
    ----------
    data: DataTable
        DataTable of figure parameters
        (sectors id, label, parent, count, proportion, p-value, ...)
    output: str
        Path to output to save the figure without extension
    analysis: str (optional, default=topology)
        Analysis mode : topology or enrichment
    test: str (optional, default=Binomial)
        Type of test for enrichment analysis : binomial or hypergeometric
    significant: Dict[str, float]
        Dictionary of significant p-value {ontology-id: p-value}
    ref_set: bool (optional, default=True)
        True if a reference set is present, False otherwise. If true will show reference set values
        in the hover text of sectors.
    write_fig: bool (optional, default=True)
        True to write the html figure, False to only return figure
    **kwargs
        Keyword args: c_min, c_max, c_mid, max_depth, colorscale, title, colorbar_legend, bg_color,
        font_color, font_size, table_title, table_legend, table_color

    Returns
    -------
    go.Figure
        Sunburst figure generated.
    """
    c_min, c_max, c_mid, max_depth, colorscale, title, colorbar_legend, background_color, \
        font_color, font_size, table_title, table_legend, table_color = \
        get_fig_kwargs(output, analysis, **kwargs)

    if analysis == TOPOLOGY_A:
        fig = go.Figure(go.Sunburst(labels=data.labels, parents=data.parents,
                                    values=data.relative_prop, ids=data.ids,
                                    hoverinfo='label+text', maxdepth=max_depth,
                                    branchvalues='total',
                                    hovertext=get_hover_fig_text(data, TOPOLOGY_A, ref_set),
                                    marker=dict(colors=data.count, colorscale=colorscale,
                                                cmin=c_min, cmax=c_max, cmid=c_mid, showscale=True,
                                                colorbar=dict(title=dict(text=colorbar_legend)))))
        fig.update_layout(title=dict(text=title, x=0.5, xanchor='center'))

    elif analysis == ENRICHMENT_A:
        fig = make_subplots(rows=1, cols=2,
                            column_widths=[0.3, 0.7],
                            vertical_spacing=0.03,
                            subplot_titles=(table_title, title),
                            specs=[[{'type': 'table'}, {'type': 'sunburst'}]])

        fig.add_trace(go.Sunburst(labels=data.labels, parents=data.parents,
                                  values=data.relative_prop, ids=data.ids,
                                  hovertext=get_hover_fig_text(data, ENRICHMENT_A, ref_set),
                                  hoverinfo='label+text', maxdepth=max_depth,
                                  branchvalues='total',
                                  marker=dict(colors=data.p_val, colorscale=colorscale,
                                              cmid=c_mid, cmax=c_max, cmin=c_min, showscale=True,
                                              colorbar=dict(title=dict(text=colorbar_legend)))),
                      row=1, col=2)

        fig.add_trace(go.Table(header=dict(values=[table_legend, f'{test} test P-value'],
                                           fill=dict(color=table_color), height=40,
                                           font=dict(size=font_size)),
                               cells=dict(values=[list(significant.keys()),
                                                  list(significant.values())],
                                          fill=dict(color=table_color), height=35,
                                          font=dict(size=font_size*0.80))),
                      row=1, col=1)
    else:
        raise ValueError('Wrong type input')
    fig.update_layout(paper_bgcolor=background_color, font_color=font_color, font_size=font_size)
    fig.update_annotations(font_size=font_size*1.5)
    if write_fig:
        fig.write_html(f'{output}.html')
    return fig


def get_hover_fig_text(data: DataTable, analysis: str, ref_set: bool) \
        -> List[str]:
    """

    Parameters
    ----------
    data
    analysis
    ref_set

    Returns
    -------

    """
    if analysis == ENRICHMENT_A:
        return [f'P value: {10 ** (-data.p_val[i])}<br>'
                f'{COUNT}: <b>{data.count[i]}</b><br>'
                f'{REF_COUNT}: {data.ref_count[i]}<br>'
                f'{PROP}: <b>{round(data.prop[i] * 100, 2)}%</b><br>'
                f'{REF_PROP}: {round(data.ref_prop[i] * 100, 2)}%<br>'
                f'{IDS}: {data.onto_ids[i]}'
                if data.p_val[i] > 0 else
                f'P value: {10 ** data.p_val[i]}<br>'
                f'{COUNT}: <b>{data.count[i]}</b><br>'
                f'{REF_COUNT}: {data.ref_count[i]}<br>'
                f'{PROP}: <b>{round(data.prop[i] * 100, 2)}%</b><br>'
                f'{REF_PROP}: {round(data.ref_prop[i] * 100, 2)}%<br>'
                f'{IDS}: {data.onto_ids[i]}'
                for i in range(data.len)]
    elif analysis == TOPOLOGY_A:
        if ref_set:
            return [f'{COUNT}: <b>{data.count[i]}</b><br>'
                    f'{REF_COUNT}: {data.ref_count[i]}<br>'
                    f'{PROP}: <b>{round(data.prop[i] * 100, 2)}%</b><br>'
                    f'{REF_PROP}: {round(data.ref_prop[i] * 100, 2)}%<br>'
                    f'{IDS}: {data.onto_ids[i]}'
                    for i in range(data.len)]
        else:
            return [f'{COUNT}: <b>{data.count[i]}</b><br>'
                    f'{PROP}: <b>{round(data.prop[i] * 100, 2)}%</b><br>'
                    f'{IDS}: {data.onto_ids[i]}'
                    for i in range(data.len)]



import inspect
import kegg2bipartitegraph
import os
import json


def update_dict(dict_1, dict_2):
    for key, value in dict_2.items():
        if key not in dict_1:
            dict_1[key] = value
        else:
            dict_1[key].extend(value)
    return dict_1


def get_parents(data_dict):
    parent_child = {}
    for key, values in data_dict.items():
        for value in values:
            if value not in parent_child:
                parent_child[value] = [key]
            else:
                parent_child[value].append(key)
        if isinstance(values, dict):
            parent_child = update_dict(parent_child, get_parents(values))
    return parent_child


k2bg_init_file = inspect.getfile(kegg2bipartitegraph)
data_path = k2bg_init_file.replace('__init__.py', 'data')
json_hierarchy = os.path.join(data_path, 'kegg_model', 'kegg_hierarchy.json')
with open(json_hierarchy, 'r') as open_json_file:
    json_data = json.load(open_json_file)


# Create dictionary for sunburst creation.
sunburst_dict = get_parents(json_data)
root = 'kegg'
for sub_root in json_data.keys():
    sunburst_dict[sub_root] = [root]

# for hierarchy in json_data:
#     sunburst_hierarchy = get_parents(json_data[hierarchy])

with open("kegg_onto.json", 'w') as open_json_file:
    json.dump(sunburst_dict, open_json_file, indent=4)


import os
import json
from typing import List, Dict
from time import time
import plotly.graph_objects as go

from ontosunburst.ontology import get_abundance_dict, get_classes_abundance, get_classes_scores, \
    extract_classes, reduce_d_ontology, METACYC, CHEBI, EC, GO, KEGG, ROOTS

from ontosunburst.data_table_tree import DataTable, get_name, BINOMIAL_TEST, ROOT_CUT, PATH_UNCUT
from ontosunburst.sunburst_fig import generate_sunburst_fig, TOPOLOGY_A, ENRICHMENT_A

# ==================================================================================================
#                                           CONSTANTS
# ==================================================================================================

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
# Dictionary json files
# ---------------------
DEFAULT_FILE = {METACYC: os.path.join(CURRENT_DIR, 'Inputs', 'MetaCyc26_0_classes.json'),
                EC: os.path.join(CURRENT_DIR, 'Inputs', 'enzymes_ontology.json'),
                KEGG: os.path.join(CURRENT_DIR, 'Inputs', 'kegg_onto.json')}
# Names json files
# ----------------
DEFAULT_NAMES = {EC: os.path.join(CURRENT_DIR, 'Inputs', 'enzymes_class_names.json'),
                 METACYC: None, KEGG: None, CHEBI: None, GO: None}
DEFAULT = 'default'
# Sparql URL
# ----------
DEFAULT_URL = {CHEBI: 'http://localhost:3030/chebi/',
               GO: 'http://localhost:3030/go/'}


# ==================================================================================================
#                                            WORKFLOW
# ==================================================================================================

def ontosunburst(interest_set: List[str],
                 ontology: str = None,
                 root: str = None,
                 abundances: List[float] = None,
                 scores: Dict[str, float] = None,
                 reference_set: List[str] = None,
                 ref_abundances: List[float] = None,
                 analysis: str = TOPOLOGY_A,
                 output: str = 'sunburst',
                 write_output: bool = True,
                 class_ontology: str or Dict[str, str] = None,
                 labels: str or Dict[str, str] = DEFAULT,
                 endpoint_url: str = None,
                 test: str = BINOMIAL_TEST,
                 root_cut: str = ROOT_CUT,
                 path_cut: str = PATH_UNCUT,
                 ref_base: bool = False,
                 show_leaves: bool = False,
                 **kwargs) -> go.Figure:
    """ Main function to be called generating the sunburst figure

    Parameters
    ----------
    interest_set: List[str]
        Interest set of concepts to classify
    ontology: str (optional, default=None, values in ['metacyc', 'ec', 'chebi', 'kegg', 'go', None])
        Ontology to use
    root: str (optional, default=None)
        Root item of the ontology.
    abundances: List[str] (optional, default=None)
        Abundance values associated to interest_set list parameter
    scores: Dict[str, float] (optional, default=None)
        Dictionary associating for each ontology ID, its enrichment score. If None enrichment will
        be calculated.
    reference_set: List[str] (optional, default=None)
        Reference set of concepts
    ref_abundances: List[str] (optional, default=None)
        Abundance values associated to reference_set list parameter
    analysis: str (optional, default='topology', values in ['topology', 'enrichment'])
        Analysis mode : topology or enrichment.
    output: str (optional, default='sunburst')
        Path of the output to save figure, if None, outputs will be sunburst.html and sunburst.tsv
        files
    write_output: bool (optional, default=True)
        True to write the html figure and tsv class files, False to only return plotly sunburst
        figure
    class_ontology: str or Dict[str, str] (optional, default=None)
        Class ontology dictionary or json file.
    labels: str or Dict[str, str] (optional, default='default')
        Path to ID-LABELS association json file or ID-LABELS association dictionary or 'default'
        to use default files. If None ontology IDs will be used as labels.
    endpoint_url: str (optional, default=None)
        URL of ChEBI or GO ontology for SPARQL requests
    test: str (optional, default='binomial', values in ['binomial', 'hypergeometric'])
        Type of test if analysis=enrichment, binomial or hypergeometric test.
    root_cut: str (optional, default='cut', values in ['uncut', 'cut', 'total'])
        mode for root cutting (uncut, cut or total)
    path_cut: str (optional, default='uncut', values in ['uncut', 'deeper', 'higher', 'bound'])
        mode for nested path cutting (uncut, deeper, higher or bound)
    ref_base: bool (optional, default=False)
        True to have the base classes representation of the reference set in the figure.
    show_leaves: bool (optional, default=False)
        True to show input metabolic objets at sunburst leaves
    **kwargs

    Returns
    -------
    go.Figure
        Plotly graph_objects figure of the sunburst
    """
    start_time = time()
    # LOAD NAMES -----------------------------------------------------------------------------------
    if labels == DEFAULT:
        if ontology is not None:
            labels = DEFAULT_NAMES[ontology]
        else:
            labels = None
    if labels is not None:
        if type(labels) == str:
            with open(labels, 'r') as f:
                names = json.load(f)
        else:
            names = labels
    else:
        names = None
    # DICTIONARY / JSON INPUT ----------------------------------------------------------------------
    if ontology == METACYC or ontology == EC or ontology == KEGG or ontology is None:
        if ontology is None:
            if class_ontology is None:
                raise ValueError('If no default ontology, must fill class_ontology parameter')
            if root is None:
                raise ValueError('If no default ontology, must fill root parameter')
        else:
            if class_ontology is None:
                class_ontology = DEFAULT_FILE[ontology]
        if type(class_ontology) == str:
            with open(class_ontology, 'r') as f:
                class_ontology = json.load(f)
    # SPARQL URL INPUT -----------------------------------------------------------------------------
    elif ontology == CHEBI or ontology == GO:
        if endpoint_url is None:
            endpoint_url = DEFAULT_URL[ontology]
    # ELSE -----------------------------------------------------------------------------------------
    else:
        raise ValueError(f'ontology parameter must be in {[METACYC, EC, KEGG, GO, CHEBI, None]}')
    # GET ROOT -------------------------------------------------------------------------------------
    if ontology is not None:
        root = ROOTS[ontology]
    # WORKFLOW -------------------------------------------------------------------------------------
    fig = _global_analysis(ontology=ontology, analysis=analysis,
                           metabolic_objects=interest_set, abundances=abundances,
                           scores=scores,
                           reference_set=reference_set, ref_abundances=ref_abundances,
                           d_classes_ontology=class_ontology, endpoint_url=endpoint_url,
                           output=output, write_output=write_output, names=names,
                           test=test, root=root, root_cut=root_cut, path_cut=path_cut,
                           ref_base=ref_base, show_leaves=show_leaves, **kwargs)
    end_time = time()
    print(f'Execution time : {end_time - start_time} seconds')
    return fig


# ==================================================================================================
#                                             FUNCTIONS
# ==================================================================================================
def _global_analysis(ontology, analysis, metabolic_objects, abundances, scores, reference_set,
                     ref_abundances, d_classes_ontology, endpoint_url, output, write_output, names,
                     test, root, root_cut, path_cut, ref_base, show_leaves, **kwargs):
    """

    Parameters
    ----------
    ontology
    analysis
    metabolic_objects
    abundances
    reference_set
    ref_abundances
    d_classes_ontology
    endpoint_url
    output
    write_output
    names
    test
    root
    root_cut
    path_cut
    ref_base
    show_leaves
    kwargs

    Returns
    -------

    """
    # EXTRACT CLASSES
    # ----------------------------------------------------------------------------------------------
    obj_all_classes, d_classes_ontology, names = extract_classes(ontology, metabolic_objects, root,
                                                                 d_classes_ontology, endpoint_url,
                                                                 names)
    if not obj_all_classes:
        print('No object classified, passing.')
    if write_output:
        write_met_classes(ontology, obj_all_classes, output, names)

    # ABUNDANCES
    # ----------------------------------------------------------------------------------------------
    # Interest
    abundances_dict = get_abundance_dict(abundances=abundances,
                                         metabolic_objects=metabolic_objects,
                                         ref=False)
    classes_abundance = get_classes_abundance(obj_all_classes, abundances_dict, show_leaves)
    # Reference
    if reference_set is not None:
        ref_abundances_dict = get_abundance_dict(abundances=ref_abundances,
                                                 metabolic_objects=reference_set,
                                                 ref=True)
        ref_all_classes, d_classes_ontology, names = extract_classes(ontology, reference_set, root,
                                                                     d_classes_ontology,
                                                                     endpoint_url, names)
        ref_classes_abundance = get_classes_abundance(ref_all_classes, ref_abundances_dict,
                                                      show_leaves)
    else:
        ref_classes_abundance = None
    # Scores
    classes_scores = None
    if scores is not None:
        classes_scores = get_classes_scores(classes_abundance, scores, root)

    # DATA TABLE
    # ----------------------------------------------------------------------------------------------
    data = DataTable()
    if ref_classes_abundance is not None:
        d_classes_ontology = reduce_d_ontology(d_classes_ontology,
                                               {**ref_classes_abundance, **classes_abundance})
        ref_set = True
        data.fill_parameters(set_abundance=classes_abundance, ref_abundance=ref_classes_abundance,
                             parent_dict=d_classes_ontology, root_item=root, names=names,
                             ref_base=ref_base)
    else:
        ref_set = False
        d_classes_ontology = reduce_d_ontology(d_classes_ontology, classes_abundance)
        data.fill_parameters(set_abundance=classes_abundance, ref_abundance=classes_abundance,
                             parent_dict=d_classes_ontology,  root_item=root, names=names,
                             ref_base=ref_base)
    data.calculate_proportions(ref_base)
    significant = None
    if analysis == ENRICHMENT_A:
        significant = data.make_enrichment_analysis(test, classes_scores)
    data.cut_root(root_cut)
    data.cut_nested_path(path_cut, ref_base)

    # FIGURE
    # ----------------------------------------------------------------------------------------------
    return generate_sunburst_fig(data=data, output=output, analysis=analysis, test=test,
                                 significant=significant, ref_set=ref_set,
                                 write_fig=write_output, **kwargs)


def write_met_classes(ontology: str, all_classes: Dict[str, List[str]], output: str,
                      names: Dict[str, str]):
    """ Writes, for each input class, all its ancestors in a .tsv file.

    Parameters
    ----------
    ontology
    all_classes
    output
    names
    """
    if ontology is None:
        ontology = ''
    links_dict = {METACYC: 'https://metacyc.org/compound?orgid=META&id=',
                  CHEBI: 'https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:',
                  EC: 'https://enzyme.expasy.org/EC/',
                  KEGG: 'https://www.genome.jp/entry/',
                  GO: 'https://amigo.geneontology.org/amigo/term/',
                  '': ''}
    with open(f'{output}.tsv', 'w') as f:
        f.write('\t'.join(['ID', 'Label', 'Classes ID', 'Classes Label', 'Link']) + '\n')
        for met_id, classes_id, in all_classes.items():
            link = links_dict[ontology] + met_id
            met_lab = get_name(met_id, names)
            classes_lab = [get_name(cl, names) for cl in classes_id]
            f.write('\t'.join([met_id, met_lab, ', '.join(classes_id), ', '.join(classes_lab),
                               link]) + '\n')



import unittest
import io
from functools import wraps

from ontosunburst.ontosunburst import *
from ontosunburst.data_table_tree import *

"""
Tests manually good file creation.
No automatic tests integrated.
"""

# ==================================================================================================
# GLOBAL
# ==================================================================================================

# Complex structure (multi-parental)
C_LST = ['a', 'b', 'c']
C_REF = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']
C_LAB = [1, 2, 3]
C_RAB = [1, 2, 3, 4, 5, 6, 7, 8]
C_ONTO = {'a': ['ab'], 'b': ['ab'], 'c': ['cde', 'cf'], 'd': ['cde'], 'e': ['cde', 'eg'],
          'f': ['cf'], 'g': ['gh', 'eg'], 'h': ['gh'],
          'ab': [ROOTS[METACYC]], 'cde': ['cdecf', 'cdeeg'], 'cf': ['cdecf'],
          'eg': [ROOTS[METACYC], 'cdeeg'], 'gh': [ROOTS[METACYC]],
          'cdecf': [ROOTS[METACYC]], 'cdeeg': ['cdeeg+'], 'cdeeg+': [ROOTS[METACYC]]}
C_LABELS = {'FRAMES': 'Root', 'cdeeg+': 'CDEEG+', 'cdeeg': 'CDEEG', 'cdecf': 'CDECF', 'gh': 'GH',
            'eg': 'EG', 'cde': 'CDE', 'cf': 'CF', 'h': 'H', 'g': 'G', 'f': 'F', 'e': 'E', 'd': 'D',
            'c': 'C', 'ab': 'AB', 'b': 'B'}

# Enrichment values
E_LST = ['02', '03', '04', '05', '08', '09']
E_LAB = [23, 20, 1, 4, 1, 1]
E_REF = ['01', '02', '03', '04', '05', '06', '07', '08', '09']
E_RAB = [14, 26, 20, 10, 20, 5, 1, 1, 3]
E_ONTO = {'01': ['00'], '02': ['00'], '03': ['00'], '04': ['00'], '05': ['01'],
          '06': ['01'], '07': ['01'], '08': ['02'], '09': ['02']}
E_LABElS = {'00': '0', '01': '1', '02': '2', '03': '3', '04': '4',
            '05': '5', '06': '6', '07': '7', '08': '8', '09': '9'}


# ==================================================================================================
# FUNCTIONS UTILS
# ==================================================================================================

def save_fig_json(fig, file):
    fig = fig.to_dict()
    with open(file, 'w') as f:
        json.dump(fig, f)


def sort_fig_dict_lst(fig):
    """
    {data: [{hovertext: [str],
             ids: [str],
             labels: [str],
             marker: {colors: [int]},
             parents: [str],
             values: [int]}]}
    """
    lst_keys = ['hovertext', 'ids', 'labels', 'parents', 'values']
    for k in lst_keys:
        fig['data'][0][k] = sorted(fig['data'][0][k])
    fig['data'][0]['marker']['colors'] = sorted(fig['data'][0]['marker']['colors'])
    return fig


def fig_to_lines(fig_dict):
    lst_keys = ['hovertext', 'ids', 'labels', 'parents', 'values']
    lines = set()
    for i in range(len(fig_dict['data'][0]['ids'])):
        line = tuple(fig_dict['data'][0][k][i] for k in lst_keys)
        line = line + (str(fig_dict['data'][0]['marker']['colors'][i]),)
        lines.add(line)
    return lines


def are_fig_dict_equals(fig1, fig2_file):
    fig1 = fig1.to_dict()
    fig1_l = fig_to_lines(fig1)
    fig1 = json.dumps(sort_fig_dict_lst(fig1), sort_keys=True)
    with open(fig2_file, 'r') as f:
        fig2 = json.load(f)
        fig2_l = fig_to_lines(fig2)
        fig2 = json.dumps(sort_fig_dict_lst(fig2), sort_keys=True)
    return (fig1 == fig2) and (fig1_l == fig2_l)


def test_for(func):
    def decorator(test_func):
        @wraps(test_func)
        def wrapper(*args, **kwargs):
            return test_func(*args, **kwargs)

        wrapper._test_for = func
        return wrapper

    return decorator


class DualWriter(io.StringIO):
    def __init__(self, original_stdout):
        super().__init__()
        self.original_stdout = original_stdout

    def write(self, s):
        super().write(s)
        self.original_stdout.write(s)


# ==================================================================================================
# UNIT TESTS
# ==================================================================================================

# CUSTOM ONTO
# --------------------------------------------------------------------------------------------------

class TestOntosunburstCustomOnto(unittest.TestCase):

    # TOPOLOGY : CUSTOM ONTO
    @test_for(ontosunburst)
    def test_ontosunburst_1(self):
        fig = ontosunburst(interest_set=C_LST, ontology=None, root='FRAMES',
                           abundances=C_LAB, reference_set=C_REF, ref_abundances=C_RAB,
                           analysis='topology', output='test1', write_output=False,
                           class_ontology=C_ONTO, labels=C_LABELS, endpoint_url=None,
                           test=BINOMIAL_TEST, root_cut=ROOT_TOTAL_CUT,
                           ref_base=True, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test1.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    @test_for(ontosunburst)
    def test_ontosunburst_2(self):
        fig = ontosunburst(interest_set=C_REF, ontology=None, root='FRAMES',
                           abundances=C_RAB, reference_set=None, ref_abundances=None,
                           analysis='topology', output='test2', write_output=False,
                           class_ontology=C_ONTO, labels=DEFAULT, endpoint_url=None,
                           test=BINOMIAL_TEST, root_cut=ROOT_CUT,
                           ref_base=False, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test2.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    @test_for(ontosunburst)
    def test_ontosunburst_3(self):
        fig = ontosunburst(interest_set=C_LST, ontology=None, root='FRAMES',
                           abundances=C_LAB, reference_set=C_REF, ref_abundances=C_RAB,
                           analysis='topology', output='test3', write_output=False,
                           class_ontology=C_ONTO, labels=C_LABELS, endpoint_url=None,
                           test=BINOMIAL_TEST, root_cut=ROOT_UNCUT,
                           ref_base=False, show_leaves=False, bg_color='#eeeeee')
        w_fig_file = os.path.join('test_files', 'test3.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    # ENRICHMENT : CUSTOM ONTO

    @test_for(ontosunburst)
    def test_ontosunburst_4(self):
        fig = ontosunburst(interest_set=E_LST, ontology=None, root='00',
                           abundances=E_LAB, reference_set=E_REF, ref_abundances=E_RAB,
                           analysis='enrichment', output='test4', write_output=False,
                           class_ontology=E_ONTO, labels=E_LABElS, endpoint_url=None,
                           test=BINOMIAL_TEST, root_cut=ROOT_CUT,
                           ref_base=True, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test4.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    @test_for(ontosunburst)
    def test_ontosunburst_5(self):
        fig = ontosunburst(interest_set=E_LST, ontology=None, root='00',
                           abundances=E_LAB, reference_set=E_REF, ref_abundances=E_RAB,
                           analysis='enrichment', output='test5', write_output=False,
                           class_ontology=E_ONTO, labels=E_LABElS, endpoint_url=None,
                           test=HYPERGEO_TEST, root_cut=ROOT_UNCUT,
                           ref_base=False, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test5.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    @test_for(ontosunburst)
    def test_ontosunburst_6(self):
        scores = {'01': 0.002, '02': 0.00007, '03': 0.9, '05': 0.004, '08': 0.2, '09': 0.0000012}
        fig = ontosunburst(interest_set=E_LST, ontology=None, root='00', scores=scores,
                           abundances=E_LAB, reference_set=E_REF, ref_abundances=E_RAB,
                           analysis='enrichment', output='test6', write_output=False,
                           class_ontology=E_ONTO, labels=E_LABElS, endpoint_url=None,
                           root_cut=ROOT_UNCUT, ref_base=False, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test6.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))


# METACYC
# --------------------------------------------------------------------------------------------------

MET_LST = ['CPD-24674', 'CPD-24687', 'CPD-24688']
MET_REF = ['CPD-24674', 'CPD-24687', 'CPD-24688',
           'CPD-12782', 'CPD-12784', 'CPD-12787',
           'CPD-12788', 'CPD-12789', 'CPD-12796',
           'CPD-12797', 'CPD-12798', 'CPD-12805',
           'CPD-12806', 'CPD-12812', 'CPD-12816',
           'CPD-1282', 'CPD-12824', 'CPD-1283']
MET_LAB = [1, 2, 3]
MET_RAB = [1, 2, 3, 4, 5, 6, 7, 8, 9, 1, 2, 3, 4, 5, 6, 7, 8, 9]

RXN_LST = ['CROTCOALIG-RXN', 'CYSTHIOCYS-RXN', 'NQOR-RXN']
RXN_REF = ['CROTCOALIG-RXN', 'CYSTHIOCYS-RXN', 'NQOR-RXN',
           'RXN-14859', 'RXN-14873', 'RXN-14920',
           'RXN-14939', 'RXN-14975', 'RXN-21632',
           'RXN-21638', 'RXN-21652', 'RXN-8954']

PWY_LST = ['2ASDEG-PWY', '4AMINOBUTMETAB-PWY', 'ALLANTOINDEG-PWY']
PWY_REF = ['2ASDEG-PWY', '4AMINOBUTMETAB-PWY', 'ALLANTOINDEG-PWY',
           'CRNFORCAT-PWY', 'PWY-7195', 'PWY-7219',
           'PWY-7251', 'PWY-7351', 'PWY-7401',
           'PWY18C3-22', 'PWY0-1600', 'SERDEG-PWY']


class TestOntosunburstMetaCyc(unittest.TestCase):

    # Topology
    @test_for(ontosunburst)
    def test_ontosunburst_mc1(self):
        fig = ontosunburst(interest_set=MET_LST, ontology=METACYC,
                           abundances=MET_LAB, reference_set=MET_REF, ref_abundances=MET_RAB,
                           analysis=TOPOLOGY_A, output='mc1', write_output=False,
                           class_ontology=None, labels=DEFAULT, endpoint_url=None,
                           test=HYPERGEO_TEST, root_cut=ROOT_CUT,
                           ref_base=True, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test_mc1.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    @test_for(ontosunburst)
    def test_ontosunburst_mc2(self):
        fig = ontosunburst(interest_set=MET_LST, ontology=METACYC,
                           abundances=MET_LAB, reference_set=MET_REF, ref_abundances=MET_RAB,
                           analysis=TOPOLOGY_A, output='mc2', write_output=False,
                           class_ontology=None, labels={'Acids': 'ACIDS !!'}, endpoint_url=None,
                           test=HYPERGEO_TEST, root_cut=ROOT_CUT,
                           ref_base=False, show_leaves=False)
        w_fig_file = os.path.join('test_files', 'test_mc2.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    @test_for(ontosunburst)
    def test_ontosunburst_mc3(self):
        fig = ontosunburst(interest_set=MET_REF, ontology=METACYC,
                           abundances=None, reference_set=None, ref_abundances=None,
                           analysis=TOPOLOGY_A, output='mc3', write_output=False,
                           class_ontology=None, labels=DEFAULT, endpoint_url=None,
                           test=HYPERGEO_TEST, root_cut=ROOT_CUT,
                           ref_base=False, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test_mc3.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    # Enrichment
    @test_for(ontosunburst)
    def test_ontosunburst_mc4(self):
        fig = ontosunburst(interest_set=MET_LST, ontology=METACYC,
                           abundances=MET_LAB, reference_set=MET_REF, ref_abundances=MET_RAB,
                           analysis=ENRICHMENT_A, output='mc4', write_output=False,
                           class_ontology=None, labels=DEFAULT, endpoint_url=None,
                           test=HYPERGEO_TEST, root_cut=ROOT_CUT,
                           ref_base=False, show_leaves=False)
        w_fig_file = os.path.join('test_files', 'test_mc4.json')
        # self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    @test_for(ontosunburst)
    def test_ontosunburst_mc5(self):
        fig = ontosunburst(interest_set=MET_LST, ontology=METACYC,
                           abundances=MET_LAB, reference_set=MET_REF, ref_abundances=MET_RAB,
                           analysis=ENRICHMENT_A, output='mc5', write_output=False,
                           class_ontology=None, labels=DEFAULT, endpoint_url=None,
                           test=BINOMIAL_TEST, root_cut=ROOT_CUT,
                           ref_base=True, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test_mc5.json')
        # self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    # Pathways
    @test_for(ontosunburst)
    def test_ontosunburst_mc6(self):
        fig = ontosunburst(interest_set=PWY_LST, ontology=METACYC,
                           abundances=None, reference_set=PWY_REF, ref_abundances=None,
                           analysis=TOPOLOGY_A, output='mc6', write_output=False,
                           class_ontology=None, labels=DEFAULT, endpoint_url=None,
                           test=BINOMIAL_TEST, root_cut=ROOT_CUT,
                           ref_base=True, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test_mc6.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    # Reactions
    @test_for(ontosunburst)
    def test_ontosunburst_mc7(self):
        fig = ontosunburst(interest_set=RXN_LST, ontology=METACYC,
                           abundances=None, reference_set=RXN_REF, ref_abundances=None,
                           analysis=TOPOLOGY_A, output='mc7', write_output=False,
                           class_ontology=None, labels=DEFAULT, endpoint_url=None,
                           test=BINOMIAL_TEST, root_cut=ROOT_CUT,
                           ref_base=True, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test_mc7.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    @test_for(ontosunburst)
    def test_ontosunburst_mc8(self):
        fig = ontosunburst(interest_set=MET_LST, ontology=METACYC,
                           abundances=MET_LAB, reference_set=MET_REF, ref_abundances=MET_RAB,
                           analysis=TOPOLOGY_A, output='mc8', write_output=False,
                           class_ontology=None, labels=DEFAULT, endpoint_url=None,
                           test=HYPERGEO_TEST, root_cut=ROOT_CUT, path_cut=PATH_BOUND,
                           ref_base=False, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test_mc8.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    @test_for(ontosunburst)
    def test_ontosunburst_mc9(self):
        fig = ontosunburst(interest_set=MET_LST, ontology=METACYC,
                           abundances=MET_LAB, reference_set=MET_REF, ref_abundances=MET_RAB,
                           analysis=TOPOLOGY_A, output='mc9', write_output=False,
                           class_ontology=None, labels=DEFAULT, endpoint_url=None,
                           test=HYPERGEO_TEST, root_cut=ROOT_CUT, path_cut=PATH_HIGHER,
                           ref_base=True, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test_mc9.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))


# EC
# --------------------------------------------------------------------------------------------------

EC_LST = ['2.6.1.45', '1.1.1.25', '1.1.1.140', '1.1.2.-']
REF_EC = ['2.6.1.45', '1.1.1.25', '1.1.1.140',
          '1.14.14.52', '2.7.1.137', '7.1.1.8',
          '1.17.4.5', '2.3.1.165', '3.2.1.53',
          '3.2.1.91', '6.3.4.2', '5.4.99.8']


class TestOntosunburstEC(unittest.TestCase):

    @test_for(ontosunburst)
    def test_ontosunburst_ec1(self):
        fig = ontosunburst(interest_set=EC_LST, ontology=EC,
                           abundances=None, reference_set=REF_EC, ref_abundances=None,
                           analysis=TOPOLOGY_A, output='ec1', write_output=False,
                           class_ontology=None, labels=DEFAULT, endpoint_url=None,
                           test=BINOMIAL_TEST, root_cut=ROOT_CUT,
                           ref_base=True, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test_ec1.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    @test_for(ontosunburst)
    def test_ontosunburst_ec2(self):
        fig = ontosunburst(interest_set=EC_LST, ontology=EC,
                           abundances=None, reference_set=REF_EC, ref_abundances=None,
                           analysis=TOPOLOGY_A, output='ec2', write_output=False,
                           class_ontology=None, labels=None, endpoint_url=None,
                           test=BINOMIAL_TEST, root_cut=ROOT_CUT,
                           ref_base=True, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test_ec2.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))

    @test_for(ontosunburst)
    def test_ontosunburst_ec3(self):
        fig = ontosunburst(interest_set=EC_LST, ontology=EC,
                           abundances=None, reference_set=REF_EC, ref_abundances=None,
                           analysis=ENRICHMENT_A, output='ec3', write_output=False,
                           class_ontology=None, labels={'1.-.-.-': ':D'}, endpoint_url=None,
                           test=BINOMIAL_TEST, root_cut=ROOT_CUT,
                           ref_base=True, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test_ec3.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))


# KEGG
# --------------------------------------------------------------------------------------------------

KEGG_EX = ['M00572', 'M00308', 'M00844', 'M00633', 'M00176', 'M00535', 'M00573', 'M00970', 'M00131',
           'M00620']


class TestOntosunburstKegg(unittest.TestCase):

    @test_for(ontosunburst)
    def test_ontosunburst_kg1(self):
        fig = ontosunburst(interest_set=KEGG_EX, ontology=KEGG,
                           abundances=None, reference_set=None, ref_abundances=None,
                           analysis=TOPOLOGY_A, output='kg1', write_output=False,
                           class_ontology=None, labels=DEFAULT, endpoint_url=None,
                           test=BINOMIAL_TEST, root_cut=ROOT_CUT,
                           ref_base=False, show_leaves=True)
        w_fig_file = os.path.join('test_files', 'test_kg1.json')
        self.assertTrue(are_fig_dict_equals(fig, w_fig_file))



import unittest
from unittest.mock import patch
import io
import sys
from functools import wraps
from ontosunburst.ontology import *

"""
Tests manually good file creation.
No automatic tests integrated.
"""

# ==================================================================================================
# GLOBAL
# ==================================================================================================

# GENERAL DICT ONTO (METACYC, KEGG)
# --------------------------------------------------------------------------------------------------

MET_LST = ['a', 'b', 'c']
MET_REF = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']
MET_LAB = [1, 2, 3]
MET_RAB = [1, 2, 3, 4, 5, 6, 7, 8]
MC_ONTO = {'a': ['ab'], 'b': ['ab'], 'c': ['cde', 'cf'], 'd': ['cde'], 'e': ['cde', 'eg'],
           'f': ['cf'], 'g': ['gh', 'eg'], 'h': ['gh'],
           'ab': [ROOTS[METACYC]], 'cde': ['cdecf', 'cdeeg'], 'cf': ['cdecf'],
           'eg': [ROOTS[METACYC], 'cdeeg'], 'gh': [ROOTS[METACYC]],
           'cdecf': [ROOTS[METACYC]], 'cdeeg': ['cdeeg+'], 'cdeeg+': [ROOTS[METACYC]]}
KG_ONTO = {'a': ['ab'], 'b': ['ab'], 'c': ['cde', 'cf'], 'd': ['cde'], 'e': ['cde', 'eg'],
           'f': ['cf'], 'g': ['gh', 'eg'], 'h': ['gh'],
           'ab': [ROOTS[KEGG]], 'cde': ['cdecf', 'cdeeg'], 'cf': ['cdecf'],
           'eg': [ROOTS[KEGG], 'cdeeg'], 'gh': [ROOTS[KEGG]],
           'cdecf': [ROOTS[KEGG]], 'cdeeg': ['cdeeg+'], 'cdeeg+': [ROOTS[KEGG]]}

# EC
# --------------------------------------------------------------------------------------------------
EC_LST = ['1.4.5.6', '1.4.6.7', '2.1.2.3', '1.5.3', '1.6.9.-', '1.-.-.-', '1.4.-.-']
EC_ONTO = {'1.4.5.-': ['1.4.-.-'], '1.4.6.-': ['1.4.-.-'], '2.1.2.-': ['2.1.-.-'],
           '1.5.3.-': ['1.5.-.-'], '1.6.9.-': ['1.6.-.-'],
           '1.4.-.-': ['1.-.-.-'], '2.1.-.-': ['2.-.-.-'], '1.5.-.-': ['1.-.-.-'],
           '1.6.-.-': ['1.-.-.-'], '1.-.-.-': [ROOTS[EC]], '2.-.-.-': [ROOTS[EC]]}
EC_ONTO_FULL = {'1.4.5.-': ['1.4.-.-'], '1.4.6.-': ['1.4.-.-'], '2.1.2.-': ['2.1.-.-'],
                '1.5.3.-': ['1.5.-.-'], '1.6.9.-': ['1.6.-.-'], '1.4.-.-': ['1.-.-.-'],
                '2.1.-.-': ['2.-.-.-'], '1.5.-.-': ['1.-.-.-'], '1.6.-.-': ['1.-.-.-'],
                '1.-.-.-': [ROOTS[EC]], '2.-.-.-': [ROOTS[EC]], '1.4.5.6': ['1.4.5.-'],
                '1.4.6.7': ['1.4.6.-'], '2.1.2.3': ['2.1.2.-'], '1.5.3': ['1.5.-.-']}


# ==================================================================================================
# FUNCTIONS UTILS
# ==================================================================================================
def dicts_with_sorted_lists_equal(dict1, dict2):
    if dict1.keys() != dict2.keys():
        return False
    for key in dict1:
        if sorted(dict1[key]) != sorted(dict2[key]):
            return False
    return True


def test_for(func):
    def decorator(test_func):
        @wraps(test_func)
        def wrapper(*args, **kwargs):
            return test_func(*args, **kwargs)

        wrapper._test_for = func
        return wrapper

    return decorator


class DualWriter(io.StringIO):
    def __init__(self, original_stdout):
        super().__init__()
        self.original_stdout = original_stdout

    def write(self, s):
        super().write(s)
        self.original_stdout.write(s)


# ==================================================================================================
# UNIT TESTS
# ==================================================================================================

# TEST CLASSES EXTRACTION
# --------------------------------------------------------------------------------------------------
class TestClassesExtraction(unittest.TestCase):
    @test_for(extract_met_classes)
    @patch('sys.stdout', new_callable=lambda: DualWriter(sys.stdout))
    def test_extract_met_classes_input_ok(self, mock_stdout):
        d_obj = extract_met_classes(MET_LST, MC_ONTO)
        output = mock_stdout.getvalue().strip()
        self.assertEqual(output, '3 metabolic objects to classify\n'
                                 '3/3 metabolic objects classified')
        self.assertEqual(d_obj, {'a': ['ab'], 'b': ['ab'], 'c': ['cde', 'cf']})

    @test_for(extract_met_classes)
    @patch('sys.stdout', new_callable=lambda: DualWriter(sys.stdout))
    def test_extract_met_classes_input_errors(self, mock_stdout):
        d_obj = extract_met_classes(MET_LST + ['x'], MC_ONTO)
        output = mock_stdout.getvalue().strip()
        self.assertEqual(output, '4 metabolic objects to classify\n'
                                 'x not classified.\n'
                                 '3/4 metabolic objects classified')
        self.assertEqual(d_obj, {'a': ['ab'], 'b': ['ab'], 'c': ['cde', 'cf']})

    @test_for(extract_ec_classes)
    @patch('sys.stdout', new_callable=lambda: DualWriter(sys.stdout))
    def test_extract_ec_classes_input_ok(self, mock_stdout):
        d_obj, d_onto = extract_ec_classes(EC_LST, EC_ONTO)
        output = mock_stdout.getvalue().strip()
        wanted_d_obj = {'1.4.5.6': ['1.4.5.-'], '1.4.6.7': ['1.4.6.-'], '2.1.2.3': ['2.1.2.-'],
                        '1.5.3': ['1.5.-.-'], '1.6.9.-': ['1.6.-.-'], '1.-.-.-': ['Enzyme'],
                        '1.4.-.-': ['1.-.-.-']}
        self.assertEqual(output, '7 EC numbers to classify\n'
                                 '7/7 EC numbers classified')
        self.assertEqual(d_obj, wanted_d_obj)
        self.assertEqual(d_onto, EC_ONTO_FULL)

    @test_for(extract_ec_classes)
    @patch('sys.stdout', new_callable=lambda: DualWriter(sys.stdout))
    def test_extract_ec_classes_input_errors(self, mock_stdout):
        d_obj, d_onto = extract_ec_classes(EC_LST + ['3.5.6.9', 'ecID'], EC_ONTO)
        output = mock_stdout.getvalue().strip()
        wanted_d_obj = {'1.4.5.6': ['1.4.5.-'], '1.4.6.7': ['1.4.6.-'], '2.1.2.3': ['2.1.2.-'],
                        '1.5.3': ['1.5.-.-'], '1.6.9.-': ['1.6.-.-'], '1.-.-.-': ['Enzyme'],
                        '1.4.-.-': ['1.-.-.-']}
        self.assertEqual(output, '9 EC numbers to classify\n'
                                 '3.5.6.9 not classified\n'
                                 'ecID not classified\n'
                                 '7/9 EC numbers classified')
        self.assertEqual(d_obj, wanted_d_obj)
        self.assertEqual(d_onto, EC_ONTO_FULL)

    @test_for(get_parents)
    def test_get_parents_linear_path(self):
        # Simple linear direction
        parents = get_parents('a', {'ab'}, MC_ONTO, ROOTS[METACYC])
        self.assertEqual(parents, {'FRAMES', 'ab'})

    @test_for(get_parents)
    def test_get_parents_complex_path(self):
        # With multiple parents having multiple parents and different size of path until root
        parents = get_parents('c', {'cde', 'cf'}, MC_ONTO, ROOTS[METACYC])
        self.assertEqual(parents, {'cdeeg+', 'FRAMES', 'cf', 'cde', 'cdecf', 'cdeeg'})

    @test_for(get_parents)
    def test_get_parents_ec(self):
        # With EC (simple path)
        parents = get_parents('2.1.2.3', {'2.1.2.-'}, EC_ONTO_FULL, ROOTS[EC])
        self.assertEqual(parents, {'Enzyme', '2.1.-.-', '2.-.-.-', '2.1.2.-'})

    @test_for(get_parents)
    def test_get_parents_direct_root_child(self):
        # With direct child of root item
        parents = get_parents('1.-.-.-', {'Enzyme'}, EC_ONTO_FULL, ROOTS[EC])
        self.assertEqual(parents, {'Enzyme'})

    @test_for(get_all_classes)
    def test_get_all_classes(self):
        leaf_classes = {'a': ['ab'], 'b': ['ab'], 'c': ['cde', 'cf']}
        all_classes_met = get_all_classes(leaf_classes, MC_ONTO, ROOTS[METACYC])
        wanted_all_classes = {'a': {'FRAMES', 'ab'}, 'b': {'FRAMES', 'ab'},
                              'c': {'cdeeg+', 'cde', 'cdeeg', 'FRAMES', 'cdecf', 'cf'}}
        self.assertEqual(all_classes_met, wanted_all_classes)

    @test_for(get_all_classes)
    def test_get_all_classes_ec(self):
        ec_leaf_classes = {'1.4.5.6': ['1.4.5.-'], '1.4.6.7': ['1.4.6.-'], '2.1.2.3': ['2.1.2.-'],
                           '1.5.3': ['1.5.-.-'], '1.6.9.-': ['1.6.-.-'], '1.-.-.-': ['Enzyme'],
                           '1.4.-.-': ['1.-.-.-']}
        wanted_all_classes = {'1.4.5.6': {'1.4.5.-', '1.4.-.-', 'Enzyme', '1.-.-.-'},
                              '1.4.6.7': {'Enzyme', '1.4.-.-', '1.4.6.-', '1.-.-.-'},
                              '2.1.2.3': {'2.-.-.-', '2.1.-.-', '2.1.2.-', 'Enzyme'},
                              '1.5.3': {'1.5.-.-', 'Enzyme', '1.-.-.-'},
                              '1.6.9.-': {'1.6.-.-', 'Enzyme', '1.-.-.-'},
                              '1.-.-.-': {'Enzyme'}, '1.4.-.-': {'Enzyme', '1.-.-.-'}}
        all_classes_ec = get_all_classes(ec_leaf_classes, EC_ONTO, ROOTS[EC])
        self.assertEqual(all_classes_ec, wanted_all_classes)

    @test_for(extract_classes)
    def test_extract_classes_metacyc(self):
        mc_classes, d_classes_ontology, names = extract_classes(METACYC, MET_LST, ROOTS[METACYC],
                                                                MC_ONTO, None)
        wanted_mc_classes = {'a': {'FRAMES', 'ab'}, 'b': {'FRAMES', 'ab'},
                             'c': {'cdeeg+', 'cde', 'cdeeg', 'FRAMES', 'cdecf', 'cf'}}
        self.assertEqual(mc_classes, wanted_mc_classes)
        self.assertTrue(dicts_with_sorted_lists_equal(d_classes_ontology, MC_ONTO))

    @test_for(extract_classes)
    def test_extract_classes_kegg(self):
        kg_classes, d_classes_ontology, names = extract_classes(KEGG, MET_LST, ROOTS[KEGG], KG_ONTO,
                                                                None)
        wanted_kg_classes = {'a': {'ab', 'kegg'}, 'b': {'ab', 'kegg'},
                             'c': {'cde', 'cdeeg+', 'kegg', 'cdecf', 'cdeeg', 'cf'}}
        self.assertEqual(kg_classes, wanted_kg_classes)
        self.assertTrue(dicts_with_sorted_lists_equal(d_classes_ontology, KG_ONTO))

    @test_for(extract_classes)
    def test_extract_classes_ec(self):
        ec_classes, d_classes_ontology, names = extract_classes(EC, EC_LST, ROOTS[EC], EC_ONTO,
                                                                None)
        wanted_ec_classes = {'1.4.5.6': {'1.4.5.-', '1.4.-.-', 'Enzyme', '1.-.-.-'},
                             '1.4.6.7': {'Enzyme', '1.4.-.-', '1.4.6.-', '1.-.-.-'},
                             '2.1.2.3': {'2.-.-.-', '2.1.-.-', '2.1.2.-', 'Enzyme'},
                             '1.5.3': {'1.5.-.-', 'Enzyme', '1.-.-.-'},
                             '1.6.9.-': {'1.6.-.-', 'Enzyme', '1.-.-.-'},
                             '1.-.-.-': {'Enzyme'}, '1.4.-.-': {'Enzyme', '1.-.-.-'}}
        self.assertEqual(ec_classes, wanted_ec_classes)
        self.assertTrue(dicts_with_sorted_lists_equal(d_classes_ontology, EC_ONTO_FULL))


# TEST ABUNDANCES
# --------------------------------------------------------------------------------------------------
class TestAbundances(unittest.TestCase):
    @test_for(get_abundance_dict)
    def test_get_abundance_dict_abundances_not_ref(self):
        abundance_dict = get_abundance_dict(abundances=MET_LAB, metabolic_objects=MET_LST,
                                            ref=False)
        self.assertEqual(abundance_dict, {'a': 1, 'b': 2, 'c': 3})

    @test_for(get_abundance_dict)
    def test_get_abundance_dict_no_abundances_not_ref(self):
        abundance_dict = get_abundance_dict(abundances=None, metabolic_objects=MET_LST,
                                            ref=False)
        self.assertEqual(abundance_dict, {'a': 1, 'b': 1, 'c': 1})

    @test_for(get_abundance_dict)
    def test_get_abundance_dict_abundances_ref(self):
        abundance_dict = get_abundance_dict(abundances=MET_RAB, metabolic_objects=MET_REF,
                                            ref=True)
        self.assertEqual(abundance_dict, {'a': 1, 'b': 2, 'c': 3, 'd': 4,
                                          'e': 5, 'f': 6, 'g': 7, 'h': 8})

    @test_for(get_abundance_dict)
    def test_get_abundance_dict_no_abundances_ref(self):
        abundance_dict = get_abundance_dict(abundances=None, metabolic_objects=MET_REF,
                                            ref=True)
        self.assertEqual(abundance_dict, {'a': 1, 'b': 1, 'c': 1, 'd': 1,
                                          'e': 1, 'f': 1, 'g': 1, 'h': 1})

    @test_for(get_abundance_dict)
    def test_get_abundance_dict_errors_not_ref(self):
        with self.assertRaises(AttributeError) as e:
            get_abundance_dict(abundances=MET_LAB + [4], metabolic_objects=MET_LST, ref=False)
        self.assertEqual(str(e.exception), 'Length of "metabolic_objects" parameter must be '
                                           'equal to "abundances" parameter length : 3 != 4')

    @test_for(get_abundance_dict)
    def test_get_abundance_dict_errors_ref(self):
        with self.assertRaises(AttributeError) as e:
            get_abundance_dict(abundances=MET_RAB[:-1], metabolic_objects=MET_REF, ref=True)
        self.assertEqual(str(e.exception), 'Length of "reference_set" parameter must be '
                                           'equal to "ref_abundances" parameter length : 8 != 7')

    @test_for(get_classes_abundance)
    def test_get_classes_abundance_leaves(self):
        all_classes = {'a': {'FRAMES', 'ab'}, 'b': {'FRAMES', 'ab'},
                       'c': {'cdecf', 'cdeeg+', 'FRAMES', 'cde', 'cdeeg', 'cf'},
                       'd': {'cdecf', 'cdeeg+', 'FRAMES', 'cde', 'cdeeg'},
                       'e': {'cdeeg+', 'FRAMES', 'cde', 'cdecf', 'eg', 'cdeeg'},
                       'f': {'cdecf', 'FRAMES', 'cf'},
                       'g': {'cdeeg', 'cdeeg+', 'FRAMES', 'eg', 'gh'}, 'h': {'FRAMES', 'gh'}}
        abundances_dict = {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5, 'f': 6, 'g': 7, 'h': 8}
        classes_abundances = get_classes_abundance(all_classes, abundances_dict, show_leaves=True)
        wanted_abundances = {'FRAMES': 36, 'cdeeg+': 19, 'cdeeg': 19, 'cdecf': 18, 'gh': 15,
                             'eg': 12, 'cde': 12, 'cf': 9, 'h': 8, 'g': 7, 'f': 6, 'e': 5,
                             'd': 4, 'c': 3, 'ab': 3, 'b': 2, 'a': 1}
        self.assertEqual(classes_abundances, wanted_abundances)

    @test_for(get_classes_abundance)
    def test_get_classes_abundance_leaves_sub(self):
        all_classes = {'a': {'FRAMES', 'ab'}, 'b': {'FRAMES', 'ab'},
                       'c': {'cdeeg+', 'cde', 'cdeeg', 'FRAMES', 'cdecf', 'cf'}}
        abundances_dict = {'a': 1, 'b': 2, 'c': 3}
        classes_abundances = get_classes_abundance(all_classes, abundances_dict, show_leaves=True)
        wanted_abundances = {'FRAMES': 6, 'cde': 3, 'cf': 3, 'cdecf': 3, 'cdeeg+': 3, 'cdeeg': 3,
                             'c': 3, 'ab': 3, 'b': 2, 'a': 1}
        self.assertEqual(classes_abundances, wanted_abundances)

    @test_for(get_classes_abundance)
    def test_get_classes_abundance_no_leaves(self):
        all_classes = {'a': {'FRAMES', 'ab'}, 'b': {'FRAMES', 'ab'},
                       'c': {'cdecf', 'cdeeg+', 'FRAMES', 'cde', 'cdeeg', 'cf'},
                       'd': {'cdecf', 'cdeeg+', 'FRAMES', 'cde', 'cdeeg'},
                       'e': {'cdeeg+', 'FRAMES', 'cde', 'cdecf', 'eg', 'cdeeg'},
                       'f': {'cdecf', 'FRAMES', 'cf'},
                       'g': {'cdeeg', 'cdeeg+', 'FRAMES', 'eg', 'gh'}, 'h': {'FRAMES', 'gh'}}
        abundances_dict = {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5, 'f': 6, 'g': 7, 'h': 8}
        classes_abundances = get_classes_abundance(all_classes, abundances_dict, show_leaves=False)
        wanted_abundances = {'FRAMES': 36, 'cdeeg+': 19, 'cdeeg': 19, 'cdecf': 18, 'gh': 15,
                             'eg': 12, 'cde': 12, 'cf': 9, 'ab': 3}
        self.assertEqual(classes_abundances, wanted_abundances)

    @test_for(get_classes_abundance)
    def test_get_classes_abundance_no_leaves_sub(self):
        all_classes = {'a': {'FRAMES', 'ab'}, 'b': {'FRAMES', 'ab'},
                       'c': {'cdeeg+', 'cde', 'cdeeg', 'FRAMES', 'cdecf', 'cf'}}
        abundances_dict = {'a': 1, 'b': 2, 'c': 3}
        classes_abundances = get_classes_abundance(all_classes, abundances_dict, show_leaves=False)
        wanted_abundances = {'FRAMES': 6, 'cde': 3, 'cf': 3, 'cdecf': 3, 'cdeeg+': 3, 'cdeeg': 3,
                             'ab': 3}
        self.assertEqual(classes_abundances, wanted_abundances)

    @test_for(get_classes_abundance)
    def test_get_classes_abundance_different_level_abundances(self):
        all_classes = {'c': {'cdecf', 'cdeeg+', 'FRAMES', 'cde', 'cdeeg', 'cf'},
                       'd': {'cdecf', 'cdeeg+', 'FRAMES', 'cde', 'cdeeg'},
                       'e': {'cdeeg+', 'FRAMES', 'cde', 'cdecf', 'eg', 'cdeeg'},
                       'f': {'cdecf', 'FRAMES', 'cf'}, 'cf': {'cdecf', 'FRAMES'}}
        abundances_dict = {'c': 3, 'd': 4, 'e': 5, 'f': 2, 'cf': 2}
        classes_abundances = get_classes_abundance(all_classes, abundances_dict, show_leaves=True)
        wanted_abundances = {'FRAMES': 16, 'cdecf': 16, 'cde': 12, 'cdeeg+': 12, 'cdeeg': 12,
                             'cf': 7, 'eg': 5, 'e': 5, 'd': 4, 'c': 3, 'f': 2}
        self.assertEqual(classes_abundances, wanted_abundances)


# TEST UTILS
# --------------------------------------------------------------------------------------------------
class TestUtils(unittest.TestCase):
    @test_for(reduce_d_ontology)
    def test_reduce_d_ontology(self):
        classes_abundance = {'FRAMES': 6, 'cde': 3, 'cf': 3, 'cdecf': 3, 'cdeeg+': 3, 'cdeeg': 3,
                             'ab': 3}
        d_ontology_reduced = reduce_d_ontology(MC_ONTO, classes_abundance)
        wanted_d_ontology_reduced = {'ab': ['FRAMES'], 'cde': ['cdecf', 'cdeeg'],
                                     'cf': ['cdecf'], 'cdecf': ['FRAMES'], 'cdeeg': ['cdeeg+'],
                                     'cdeeg+': ['FRAMES']}
        self.assertEqual(d_ontology_reduced, wanted_d_ontology_reduced)





from typing import List, Set, Dict, Tuple

import numpy
from SPARQLWrapper import SPARQLWrapper, JSON


# CONSTANTS ========================================================================================


METACYC = 'metacyc'
EC = 'ec'
CHEBI = 'chebi'
GO = 'go'
KEGG = 'kegg'

ROOTS = {METACYC: 'FRAMES',
         CHEBI: 'role',
         EC: 'Enzyme',
         GO: 'GO',
         KEGG: 'kegg'}

GO_ROOTS = ['cellular_component', 'biological_process', 'molecular_function']


# ==================================================================================================
# CLASSES EXTRACTION
# ==================================================================================================

# Main class extraction function
# --------------------------------------------------------------------------------------------------
def extract_classes(ontology: str, concepts: List[str], root: str,
                    d_classes_ontology: Dict[str, List[str]] = None, endpoint_url: str = None,
                    names: Dict[str, str] = None)\
        -> Tuple[Dict[str, Set[str]], Dict[str, List[str]], Dict[str, str] or None]:
    """ Extract all parent classes (until root) from a list of concepts.

    Parameters
    ----------
    ontology: str
        Name of the ontology considered
    concepts: List[str]
        List of concepts to classify
    root: str
        Root item of the ontology
    d_classes_ontology: Dict[str, List[str]], optional (default=None)
        Dictionary of the classes ontology associating for each concept its +1 parent classes.
    endpoint_url: str, optional (default=None)
        URL for the SPARQL server (for GO and ChEBI ontologies)
    names: Dict[str, str] (default=None)

    Returns
    -------
    Dict[str, Set[str]]
        Dictionary associating to each concept all its parent classes (until the root)
    Dict[str, List[str]]
        Dictionary of the classes ontology associating for each concept its +1 parent classes.
    Dict[str, str]  or None
        Dictionary of labels
    """
    if ontology == METACYC or ontology == KEGG or ontology is None:
        leaf_classes = extract_met_classes(concepts, d_classes_ontology)
        return get_all_classes(leaf_classes, d_classes_ontology, root), d_classes_ontology, names
    if ontology == EC:
        leaf_classes, d_classes_ontology = extract_ec_classes(concepts, d_classes_ontology)
        return get_all_classes(leaf_classes, d_classes_ontology, root), d_classes_ontology, names
    if ontology == CHEBI:
        return extract_chebi_roles(concepts, endpoint_url)
    if ontology == GO:
        return extract_go_classes(concepts, endpoint_url)


# For MetaCyc and Kegg Ontology
# --------------------------------------------------------------------------------------------------
def extract_met_classes(concepts: List[str], d_classes_ontology: Dict[str, List[str]]) \
        -> Dict[str, List[str]]:
    """ Extract +1 parent classes for each concept considered.

    Parameters
    ----------
    concepts: List[str]
        List of concepts considered
    d_classes_ontology: Dict[str, List[str]]
        Dictionary of the classes ontology associating for each concept its +1 parent classes.

    Returns
    -------
    Dict[str, List[str]]
        Dictionary associating for each concept, the list of +1 parent classes it belongs to.
    """
    d_obj_classes = dict()
    print(f'{len(concepts)} metabolic objects to classify')
    for obj in concepts:
        try:
            d_obj_classes[obj] = d_classes_ontology[obj]
            classified = True
        except KeyError:
            classified = False
        if not classified:
            print(f'{obj} not classified.')
    print(f'{len(d_obj_classes)}/{len(concepts)} metabolic objects classified')
    return d_obj_classes


# For EC
# --------------------------------------------------------------------------------------------------
def extract_ec_classes(ec_list: List[str], d_classes_ontology: Dict[str, List[str]]) \
        -> Tuple[Dict[str, List[str]], Dict[str, List[str]]]:
    """ Extract +1 parent classes for each EC number.

    Parameters
    ----------
    ec_list: List[str]
        List of EC numbers
    d_classes_ontology: Dict[str, List[str]]
        EC Ontology classes dictionary : Dict[object, List[parents]]

    Returns
    -------
    Dict[str, List[str]]
        Dictionary associating for each metabolic object, the list of +1 parent classes it belongs
        to.
    Dict[str, List[str]]
        EC Ontology classes dictionary : Dict[object, List[parents]]
    """
    print(f'{len(ec_list)} EC numbers to classify')
    ec_classes = dict()
    for ec in ec_list:
        dec_ec = ec.split('.')
        while len(dec_ec) < 4:
            dec_ec.append('-')
        if dec_ec.count('-') == 3 and ec in d_classes_ontology:
            parent = ROOTS[EC]
        else:
            for i in range(3):
                if dec_ec.count('-') == i:
                    dec_ec[3-i] = '-'
                    break
            parent = '.'.join(dec_ec)
        if parent in d_classes_ontology or parent == ROOTS[EC]:
            d_classes_ontology[ec] = [parent]
            ec_classes[ec] = [parent]
        else:
            print(f'{ec} not classified')
    print(f'{len(ec_classes)}/{len(ec_list)} EC numbers classified')
    return ec_classes, d_classes_ontology


# For ChEBI Ontology
# --------------------------------------------------------------------------------------------------
def extract_chebi_roles(chebi_ids: List[str], endpoint_url: str) \
        -> Tuple[Dict[str, Set[str]], Dict[str, List[str]], Dict[str, str]]:
    """ Extract all parent classes for each chebi ID + Generate ontology dictionary.

    Parameters
    ----------
    chebi_ids: List[str]
        List of ChEBI IDs to extract roles associated
    endpoint_url: str
        URL endpoint string

    Returns
    -------
    Tuple[Dict[str, Set[str]], Dict[str, List[str]], Dict[str, str]]
        Dictionary of all roles associated for each ChEBI ID
        Dictionary of roles ontology, associating for each role, its parent roles
        Dictionary of labels
    """
    root_id = '50906'
    d_roles_ontology = dict()
    all_roles = dict()
    d_labels = dict()
    chebi_ok = 0
    total_nb = len(chebi_ids)
    for chebi_id in chebi_ids:
        roles = set()
        sparql = SPARQLWrapper(endpoint_url)
        sparql.setQuery(f"""
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
        PREFIX owl: <http://www.w3.org/2002/07/owl#>
        PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
        PREFIX dc: <http://purl.org/dc/elements/1.1/>
        PREFIX dcterms: <http://purl.org/dc/terms/>
        PREFIX chebi: <http://purl.obolibrary.org/obo/chebi/>
        PREFIX chebidb: <http://purl.obolibrary.org/obo/CHEBI_>
        PREFIX chebirel: <http://purl.obolibrary.org/obo/chebi#>
        PREFIX obo: <http://purl.obolibrary.org/obo/>
        PREFIX oboInOwl: <http://www.geneontology.org/formats/oboInOwl#>
        PREFIX bp3: <http://www.biopax.org/release/biopax-level3.owl#>
    
        SELECT DISTINCT ?molecule ?moleculeLabel ?roleId ?roleLabel ?parentRoleLabel ?parentRoleId
        WHERE {{
            VALUES ?molecule{{ chebidb:{chebi_id}}}                                        
            
            ?molecule rdfs:label ?moleculeLabel .
            
            ?molecule rdfs:subClassOf+ ?restriction .
            ?restriction rdf:type owl:Restriction .
            ?restriction owl:onProperty obo:RO_0000087 .
            ?restriction owl:someValuesFrom/(rdfs:subClassOf*) ?role .
            
            ?role oboInOwl:id ?roleId .
            ?role rdfs:subClassOf ?parentRole .
            
            ?parentRole oboInOwl:id ?parentRoleId .
            ?parentRole rdfs:label ?parentRoleLabel .
            ?role rdfs:label ?roleLabel .
        }}
        """)
        sparql.setReturnFormat(JSON)
        results = sparql.query().convert()
        parent_roles = set()
        for result in results["results"]["bindings"]:
            molecule_label = result['moleculeLabel']['value']
            role_label = result['roleLabel']['value']
            role_id = result['roleId']['value'].split(':')[1]
            parent_role_label = result['parentRoleLabel']['value']
            parent_role_id = result['parentRoleId']['value'].split(':')[1]
            if role_id == root_id:
                role_id = ROOTS[CHEBI]
            if parent_role_id == root_id:
                parent_role_id = ROOTS[CHEBI]
            d_labels[role_id] = role_label
            d_labels[parent_role_id] = parent_role_label
            d_labels[chebi_id] = molecule_label
            parent_roles.add(parent_role_id)
            roles.add(role_id)
            if role_id not in d_roles_ontology:
                d_roles_ontology[role_id] = set()
            d_roles_ontology[role_id].add(parent_role_id)
        if roles:
            chebi_ok += 1
            d_roles_ontology[chebi_id] = list(roles.difference(parent_roles))
            roles.add(ROOTS[CHEBI])
            all_roles[chebi_id] = roles
        else:
            print(f'No ChEBI role found for : {chebi_id}')

    for c, p in d_roles_ontology.items():
        d_roles_ontology[c] = list(p)
    print(f'{chebi_ok}/{total_nb} chebi id with roles associated.')
    return all_roles, d_roles_ontology, d_labels


# For GO Ontology
# --------------------------------------------------------------------------------------------------
def extract_go_classes(go_ids: List[str], endpoint_url: str) \
        -> Tuple[Dict[str, Set[str]], Dict[str, List[str]], Dict[str, str]]:
    """ Extract all parent classes for each GO ID + Generate ontology dictionary.

    Parameters
    ----------
    go_ids: List[str]
        List of GO IDs
    endpoint_url: str
        URL endpoint string

    Returns
    -------
    Tuple[Dict[str, Set[str]], Dict[str, List[str]], Dict[str, str]]
        Dictionary of all roles associated for each ChEBI ID
        Dictionary of roles ontology, associating for each role, its parent roles
        Dictionary of labels
    """
    d_classes_ontology = dict()
    all_classes = dict()
    d_labels = dict()
    for go in go_ids:
        go_classes = set()
        sparql = SPARQLWrapper(endpoint_url)
        sparql.setQuery(f"""
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX rdfs:<http://www.w3.org/2000/01/rdf-schema#>
        PREFIX owl: <http://www.w3.org/2002/07/owl#>
        PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
        PREFIX dc: <http://purl.org/dc/elements/1.1/>
        PREFIX dcterms: <http://purl.org/dc/terms/>
        PREFIX oboInOwl: <http://www.geneontology.org/formats/oboInOwl#>
        PREFIX taxon: <http://purl.uniprot.org/taxonomy/>
        PREFIX uniprot: <http://purl.uniprot.org/uniprot/>
        PREFIX up:<http://purl.uniprot.org/core/>
        PREFIX go: <http://purl.obolibrary.org/obo/GO_>
        PREFIX goavoc: <http://bio2rdf.org/goa_vocabulary:>
        
        SELECT ?goLabel ?parentGoLabel ?goId ?parentGoId
        WHERE {{
           {go} rdfs:subClassOf* ?go .
           ?go oboInOwl:id ?goId .
           ?go rdfs:label ?goLabel .
           ?go rdf:type owl:Class .
           ?go rdfs:subClassOf ?parentGo .
           ?parentGo rdfs:label ?parentGoLabel .
           ?parentGo oboInOwl:id ?parentGoId .
           ?parentGo rdf:type owl:Class .
        }}
        """)
        sparql.setReturnFormat(JSON)
        results = sparql.query().convert()
        for result in results["results"]["bindings"]:
            go_id = result['goId']['value'].lower()
            go_label = result['goLabel']['value']
            parent_id = result['parentGoId']['value'].lower()
            parent_label = result['parentGoLabel']['value']
            d_labels[go_id] = go_label
            d_labels[parent_id] = parent_label
            go_classes.add(parent_id)
            if parent_label in GO_ROOTS:
                d_classes_ontology[parent_id] = [ROOTS[GO]]
            if go_id not in d_classes_ontology:
                d_classes_ontology[go_id] = []
            d_classes_ontology[go_id].append(parent_id)
        if go_classes:
            go_classes.add(ROOTS[GO])
            all_classes[go] = go_classes
        else:
            print(f'No GO class found for : {go}')

    for c, p in d_classes_ontology.items():
        d_classes_ontology[c] = list(set(p))
    return all_classes, d_classes_ontology, d_labels


# Recursive class extraction function
# --------------------------------------------------------------------------------------------------
def get_all_classes(obj_classes: Dict[str, List[str]], d_classes_ontology: Dict[str, List[str]],
                    root_item: str) -> Dict[str, Set[str]]:
    """ Extract all parent classes for each metabolite.

    Parameters
    ----------
    obj_classes: Dict[str, List[str]] (Dict[metabolite, List[class]])
        Dictionary associating for each object the list of +1 parent classes it belongs to.
    d_classes_ontology: Dict[str, List[str]]
        Dictionary of the classes ontology associating for each class its +1 parent classes.
    root_item: str
        Name of the root item of the ontology.

    Returns
    -------
    Dict[str, Set[str]] (Dict[metabolite, Set[class]])
        Dictionary associating for each metabolite the list of all parent classes it belongs to.
    """
    all_classes_met = dict()
    for met, classes in obj_classes.items():
        all_classes = set(classes)
        for c in classes:
            if c != root_item:
                m_classes = get_parents(c, set(d_classes_ontology[c]), d_classes_ontology, root_item)
                all_classes = all_classes.union(m_classes)
        all_classes_met[met] = all_classes
    return all_classes_met


def get_parents(child: str, parent_set: Set[str], d_classes_ontology: Dict[str, List[str]],
                root_item) -> Set[str]:
    """ Get recursively from a child class, all its parents classes found in ontology.

    Parameters
    ----------
    child: str
        Child class
    parent_set: Set[str]
        Set of all parents from previous classes
    d_classes_ontology: Dict[str, List[str]]
        Dictionary of the classes ontology of MetaCyc associating for each class its parent classes.
    root_item: str
        Name of the root item of the ontology

    Returns
    -------
    Set[str]
        Set of the union of the set  of child parent classes and the set of all previous parents.
    """
    parents = d_classes_ontology[child]
    for p in parents:
        parent_set.add(p)
    for p in parents:
        if p != root_item:
            parent_set = get_parents(p, parent_set, d_classes_ontology, root_item)
    return parent_set


# ==================================================================================================
# ABUNDANCES CALCULATION
# ==================================================================================================

def get_abundance_dict(abundances: List[float] or None, metabolic_objects: List[str], ref: bool)\
        -> Dict[str, float]:
    """ Generate abundances dictionary.

    Parameters
    ----------
    abundances: List[float] (size N) or None
        List of metabolic objects abundances (or None if no abundances associated --> will associate
        an abundance of 1 for each object)
    metabolic_objects: List[str] (size N)
        List of metabolic objects ID.
    ref: bool
        True if metabolic objects are the reference list, False otherwise (subset / study case).

    Returns
    -------
    Dict[str, float]
        Dictionary associating to each object its abundance.
    """
    if abundances is None:
        abundances = len(metabolic_objects) * [1]
    if len(metabolic_objects) == len(abundances):
        abundances_dict = {}
        for i in range(len(metabolic_objects)):
            abundances_dict[metabolic_objects[i]] = abundances[i]
    else:
        if ref:
            raise AttributeError(f'Length of "reference_set" parameter must be equal to '
                                 f'"ref_abundances" parameter length : {len(metabolic_objects)} '
                                 f'!= {len(abundances)}')
        else:
            raise AttributeError(f'Length of "metabolic_objects" parameter must be equal to '
                                 f'"abundances" parameter length : {len(metabolic_objects)} '
                                 f'!= {len(abundances)}')
    return abundances_dict


def get_classes_abundance(all_classes: Dict[str, Set[str]], abundances_dict: Dict[str, float],
                          show_leaves: bool) -> Dict[str, float]:
    """ Indicate for each class the number of base object found belonging to the class

    Parameters
    ----------
    all_classes: Dict[str, Set[str]] (Dict[metabolite, Set[class]])
        Dictionary associating for each concept the list of all parent classes it belongs to.
    abundances_dict: Dict[str, float]
        Dictionary associating for each concept, its abundance value
    show_leaves: bool
        True to show input metabolic objets at sunburst leaves

    Returns
    -------
    Dict[str, float]
        Dictionary associating for each class the weight of concepts found belonging to the class.
    """
    classes_abondance = dict()
    for met, classes in all_classes.items():
        if show_leaves:
            if met not in classes_abondance.keys():
                classes_abondance[met] = abundances_dict[met]
            else:
                classes_abondance[met] += abundances_dict[met]
        for c in classes:
            if c not in classes_abondance.keys():
                classes_abondance[c] = abundances_dict[met]
            else:
                classes_abondance[c] += abundances_dict[met]
    return dict(reversed(sorted(classes_abondance.items(), key=lambda item: item[1])))


def get_classes_scores(all_classes, scores_dict, root):
    classes_scores = dict()
    for met, classes in all_classes.items():
        if met in scores_dict.keys():
            classes_scores[met] = scores_dict[met]
        else:
            classes_scores[met] = numpy.nan
    classes_scores[root] = numpy.nan
    return classes_scores


# ==================================================================================================
# UTILS
# ==================================================================================================

def reduce_d_ontology(d_classes_ontology: Dict[str, List[str]],
                      classes_abundance: Dict[str, float]) -> Dict[str, List[str]]:
    """ Extract the sub-graph of the d_classes_ontology dictionary conserving only nodes implicated
    with the concepts studied.

    Parameters
    ----------
    d_classes_ontology: Dict[str, List[str]]
        Dictionary of the ontology complete graph
    classes_abundance: Dict[str, float]
        Dictionary of abundances (keys are all nodes implicated to be conserved)

    Returns
    -------
    Dict[str, List[str]]
        Dictionary of the ontology sub-graph conserving only nodes implicated with the concepts
        studied.
    """
    reduced_d_ontology = dict()
    for k, v in d_classes_ontology.items():
        if k in classes_abundance:
            reduced_d_ontology[k] = v
    return reduced_d_ontology



import unittest
from ontosunburst.ontosunburst import *
"""
Tests manually good file creation.
No automatic tests integrated.
"""

# METACYC
# ==================================================================================================

MET_LST = ['CPD-24674', 'CPD-24687', 'CPD-24688']
MET_REF = ['CPD-24674', 'CPD-24687', 'CPD-24688',
           'CPD-12782', 'CPD-12784', 'CPD-12787',
           'CPD-12788', 'CPD-12789', 'CPD-12796',
           'CPD-12797', 'CPD-12798', 'CPD-12805',
           'CPD-12806', 'CPD-12812', 'CPD-12816',
           'CPD-1282', 'CPD-12824', 'CPD-1283']
MET_LAB = [1, 2, 3]
MET_RAB = [1, 2, 3, 4, 5, 6, 7, 8, 9, 1, 2, 3, 4, 5, 6, 7, 8, 9]

RXN_LST = ['CROTCOALIG-RXN', 'CYSTHIOCYS-RXN', 'NQOR-RXN']
RXN_REF = ['CROTCOALIG-RXN', 'CYSTHIOCYS-RXN', 'NQOR-RXN',
           'RXN-14859', 'RXN-14873', 'RXN-14920',
           'RXN-14939', 'RXN-14975', 'RXN-21632',
           'RXN-21638', 'RXN-21652', 'RXN-8954']

PWY_LST = ['2ASDEG-PWY', '4AMINOBUTMETAB-PWY', 'ALLANTOINDEG-PWY']
PWY_REF = ['2ASDEG-PWY', '4AMINOBUTMETAB-PWY', 'ALLANTOINDEG-PWY',
           'CRNFORCAT-PWY', 'PWY-7195', 'PWY-7219',
           'PWY-7251', 'PWY-7351', 'PWY-7401',
           'PWY18C3-22', 'PWY0-1600', 'SERDEG-PWY']


class MetacycTest(unittest.TestCase):
    # Compounds
    def test_cpd_metacyc_proportion(self):
        fig = ontosunburst(ontology=METACYC, interest_set=MET_LST,
                           output='test_mc_cpd_prop', ref_base=False, abundances=MET_LAB,
                           show_leaves=True)
        fig.write_image('test_mc_cpd_prop.png', width=1900, height=1000, scale=1)

    def test_cpd_metacyc_comparison(self):
        fig = ontosunburst(ontology=METACYC, interest_set=MET_LST, reference_set=MET_REF,
                           analysis=ENRICHMENT_A, output='test_mc_cpd_comp', ref_base=True)
        fig.write_image('test_mc_cpd_comp.png', width=1900, height=1000, scale=1)

    # Reactions
    def test_rxn_metacyc_proportion(self):
        fig = ontosunburst(ontology=METACYC, interest_set=RXN_LST, reference_set=RXN_REF,
                           output='test_mc_rxn_prop', ref_base=True)
        fig.write_image('test_mc_rxn_prop.png', width=1900, height=1000, scale=1)

    def test_rxn_metacyc_comparison(self):
        fig = ontosunburst(ontology=METACYC, interest_set=RXN_LST, reference_set=RXN_REF,
                           analysis=ENRICHMENT_A, output='test_mc_rxn_comp', ref_base=True)
        fig.write_image('test_mc_rxn_comp.png', width=1900, height=1000, scale=1)

    # Pathways
    def test_pwy_metacyc_proportion(self):
        fig = ontosunburst(ontology=METACYC, interest_set=PWY_LST, reference_set=PWY_REF,
                           output='test_mc_pwy_prop', ref_base=True)
        fig.write_image('test_mc_pwy_prop.png', width=1900, height=1000, scale=1)

    def test_pwy_metacyc_comparison(self):
        fig = ontosunburst(ontology=METACYC, interest_set=PWY_LST, reference_set=PWY_REF,
                           analysis=ENRICHMENT_A, output='test_mc_pwy_comp', ref_base=True)
        fig.write_image('test_mc_pwy_comp.png', width=1900, height=1000, scale=1)


# EC
# ==================================================================================================

EC_LST = ['2.6.1.45', '1.1.1.25', '1.1.1.140', '1.1.2.-']
REF_EC = ['2.6.1.45', '1.1.1.25', '1.1.1.140',
          '1.14.14.52', '2.7.1.137', '7.1.1.8',
          '1.17.4.5', '2.3.1.165', '3.2.1.53',
          '3.2.1.91', '6.3.4.2', '5.4.99.8']


class EcTest(unittest.TestCase):

    def test_ec_proportion(self):
        fig = ontosunburst(ontology=EC, interest_set=EC_LST, reference_set=REF_EC,
                           output='test_ec_prop', ref_base=False, show_leaves=True)
        fig.write_image('test_ec_prop.png', width=1900, height=1000, scale=1)

    def test_ec_comparison(self):
        fig = ontosunburst(ontology=EC, interest_set=EC_LST, reference_set=REF_EC,
                           output='test_ec_comp', analysis=ENRICHMENT_A, ref_base=True)
        fig.write_image('test_ec_comp.png', width=1900, height=1000, scale=1)


# CHEBI
# ==================================================================================================

URL = 'http://localhost:3030/chebi/'
CH_LST = ['38028', '28604', '85146']
REF_CH = ['38028', '28604', '85146',
          '23066', '27803', '37565',
          '58215', '79983', '42639']


class ChEbiTest(unittest.TestCase):

    def test_chebi_proportion(self):
        fig = ontosunburst(ontology=CHEBI, interest_set=CH_LST, reference_set=REF_CH,
                           endpoint_url=URL, output='test_chebi_prop', ref_base=True)
        fig.write_image('test_chebi_prop.png', width=1900, height=1000, scale=1)

    def test_chebi_comparison(self):
        fig = ontosunburst(ontology=CHEBI, interest_set=CH_LST, reference_set=REF_CH,
                           endpoint_url=URL, output='test_chebi_comp', analysis=ENRICHMENT_A,
                           ref_base=True)
        fig.write_image('test_chebi_comp.png', width=1900, height=1000, scale=1)


# GO
# ==================================================================================================

GO_EX = ['go:0043226', 'go:0043227', 'go:0043229', 'go:0043231', 'go:0044422', 'go:0044424',
         'go:0044429', 'go:0044444']


class GOTest(unittest.TestCase):

    def test_go_proportion(self):
        fig = ontosunburst(ontology=GO, interest_set=GO_EX,
                           output='test_go_prop', ref_base=False, show_leaves=True)
        fig.write_image('test_go_prop.png', width=1900, height=1000, scale=1)


# KEGG
# ==================================================================================================

KEGG_EX = ['M00572', 'M00308', 'M00844', 'M00633', 'M00176', 'M00535', 'M00573', 'M00970', 'M00131',
           'M00620']


class KeggTest(unittest.TestCase):

    def test_kegg_proportion(self):
        fig = ontosunburst(ontology=GO, interest_set=GO_EX,
                           output='test_go_prop', ref_base=False, show_leaves=True)
        fig.write_image('test_kegg_prop.png', width=1900, height=1000, scale=1)




cl_tst1 = 'ontosunburst -i tests/global_test/test1_int.tsv -ir tests/global_test/test1_ref.tsv ' \
          '--onto metacyc --pcut deeper --kwargs bg_color=pink title="C\'est un titre" mx_dpt=3'

cl_tst2 = ' ontosunburst -i Files/test2_int.tsv -ir Files/test1_ref.tsv --onto metacyc ' \
          '-a enrichment -o test2 -sl --pcut deeper'



import json
import os.path
from padmet.classes.padmetRef import PadmetRef

METACYC_PADMET = os.path.join('Data/metacyc_26.0_prot70.padmet')


def metacyc_file(input_padmet=METACYC_PADMET, output_name='MetaCyc', version='26_0'):
    output = output_name + version + '_classes.json'
    pref = PadmetRef(input_padmet)
    rels = pref.getAllRelation()
    classes = dict()
    for r in rels:
        if r.type == 'is_a_class':
            if r.id_in not in classes:
                classes[r.id_in] = set()
            classes[r.id_in].add(r.id_out)
    for c, p in classes.items():
        classes[c] = list(p)
    with open(output, 'w') as o:
        json.dump(fp=o, obj=classes)


metacyc_file()
